/* dhcp4_lease.c
 *
 * Convert the ISC dhcp lease options into libdhcp option tree
 *
 * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of
 * the GNU General Public License v.2.  This program is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY expressed or
 * implied, including the implied warranties of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.  You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA. Any Red Hat trademarks that are incorporated in the source code or
 * documentation are not subject to the GNU General Public License and may
 * only be used or replicated with the express permission of Red Hat, Inc.
 *
 * Red Hat Author(s): Jason Vas Dias
 *                    David Cantrell
 */

#include <sys/types.h>
#include <unistd.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <time.h>
#include <search.h>
extern void tdestroy (void *root, void (*free_node)(void *nodep));
#include <stdio.h>

#include <isc_dhcp/dhcpd.h>

#include <dhcp4_lease.h>

static int option_code_comparator( const void *p1, const void *p2 )
{    
    const DHCPv4_option *o1 = p1, *o2 = p2;
    uint16_t o1code = (o1->unicode << 8) | (o1->code),
 	     o2code = (o2->unicode << 8) | (o2->code);
    return
	(  ( o1code == o2code ) 
	   ? 0
	   :( ( o1code > o2code )
	      ? 1
	      : -1
	    )
	);
}

static int option_name_comparator( const void *p1, const void *p2 )
{
    const DHCPv4_option *o1 = p1, *o2 = p2;
    return strcmp(o1->name, o2->name);
}

DHCPv4_option *dhcpv4_get_option_by_code( DHCPv4_lease *l, uint8_t universe, uint8_t code)
{
    DHCPv4_option sop, **opp;
    sop.unicode = universe;
    sop.code = code;
    if( (opp = tfind( &sop, &(l->options), option_code_comparator)) != 0L )
	return *opp;
    return 0L;
}

DHCPv4_option *dhcpv4_get_option_by_name( DHCPv4_lease *l, char *n)
{
    DHCPv4_option sop, **opp;
    sop.name = n;
    if( (opp = tfind( &sop, &(l->options_by_name), option_name_comparator)) != 0L )
	return *opp;
    return 0L;
}

static 
void dhcp4_client_option_add 
(   struct option_cache *oc,
    struct packet *packet, 
    struct lease *isc_lease,
    struct client_state *client,
    struct option_state *in_options,
    struct option_state *cfg_options,
    struct binding_scope **scope,
    struct universe *u, 
    void   *lease_ptr
);

DHCPv4_lease *dhcpv4_lease( struct client_state *client )
{
    DHCPv4_lease *lease;
    struct client_lease *l = client->new ? client->new : client->active;
    int i;

    if ((l == 0L) || (l->address.len != 4))
        return 0L;
    
    lease = calloc(1, sizeof(DHCPv4_lease));

    lease->address.s_addr = 
        htonl(  (l->address.iabuf[0] << 24)
              | (l->address.iabuf[1] << 16)
              | (l->address.iabuf[2] << 8)
              | l->address.iabuf[3]);

    lease->requested_address.s_addr = 
        htonl( (client->requested_address.iabuf[0] << 24)
              | (client->requested_address.iabuf[1] << 16)
              | (client->requested_address.iabuf[2] << 8)
              | client->requested_address.iabuf[3]);

    lease->is_static = l->is_static;
    lease->is_bootp = l->is_bootp;
    lease->requested = client->first_sending;
    lease->expiry = l->expiry;
    lease->renewal = l->renewal;
    lease->rebind = l->rebind;

    if (client->interface) {
        lease->if_index = client->interface->index;    
        if (client->interface->name)
            lease->if_name = strdup(client->interface->name);
    } else {
        lease->if_index = -1;
    }

    lease->server_address = client->packet.siaddr;
    
    if ( l->filename )
        lease->filename = strdup(l->filename);
    
    if (l->server_name)
        lease->server_name = strdup(l->server_name);

    for (i = 0; i < l -> options -> universe_count; i++) {
        /*fprintf(stderr,"UNIVERSE: %d %s\n", universes[i]->index, universes[i]->name);*/
        option_space_foreach((struct packet *)0, (struct lease *)0,
                             client, (struct option_state *)0,
                             l->options, &global_scope, universes[i],
                             lease, dhcp4_client_option_add);
    }

    return lease;
}

static
void option_free( void *opp )
{
    DHCPv4_option *opt = opp;
    free(opt->name);
    free(opt->format);
    free(opt->universe);
    free(opt);
}

void dhcpv4_lease_free( DHCPv4_lease *lease)
{
    if (lease->if_name) {
        free(lease->if_name);
        lease->if_name = 0;
    }

    if (lease->filename) {
        free(lease->filename);
        lease->filename = 0;
    }

    if (lease->server_name) {
        free(lease->server_name);
        lease->server_name=0;
    }

    if (lease->options) {
        tdestroy(lease->options, option_free);
        lease->options = 0;
    }

    free(lease);
}

static void option_twalker( const void *p, const VISIT which, const int depth )
{
    DHCPv4_option *option, *const*opp=p;
    DHCPv4_lease  *lease;

    if( (opp == 0L) || ((option = *opp) == 0L)
      ||( (which != postorder) && (which != leaf) )
      ) return;
    
    lease = option -> lease;
    
    lease -> handler (option, lease -> handler_arg );
}

void dhcpv4_process_options ( DHCPv4_lease *lease, DHCPv4_option_handler handler, void *handler_arg)
{    
    if ( lease &&  lease->options )
    {
	lease->handler = handler;
	lease->handler_arg = handler_arg;
	twalk(lease->options, option_twalker);
    }
} 

#define DHC_PAD( type, total_length ) ( ( sizeof(type) - (total_length  & (sizeof(type)-1) ) ) & (sizeof(type)-1) )

void dhcp4_client_option_add 
(   struct option_cache *oc,
    struct packet *packet, 
    struct lease *isc_lease,
    struct client_state *client,
    struct option_state *in_options,
    struct option_state *cfg_options,
    struct binding_scope **scope,
    struct universe *u, 
    void   *lease_ptr
)
{
    DHCPv4_lease *lease = lease_ptr;
    DHCPv4_option *option;
    struct data_string data;
    const  uint8_t *dp, *de;
    uint8_t *p, *value, *endv, 
	    is_array = 0, element_pad = 0, **vl;
    int fi=0, i=0, length = 0, n_values = 0, total_length = 0, dbg_length=0;
    uint32_t intv;
    uint16_t sv, n_members=0, n_elements =0, member_size=0;
    const uint8_t *dbg_d_last;
    uint8_t *dbg_v_last, **dbg_vl_last;
    void *dbg_t;
    
/*   fprintf(stderr,"dhcp4_client_option_add: %s\n", oc->option->name);*/

    memset (&data, 0, sizeof data);
    if ( evaluate_option_cache 
	 (   &data, packet, isc_lease, client,
	     in_options, cfg_options, scope, oc, MDL
	 ) 
       ) 
    {
	if (data.len) 
	{

/*	    fprintf(stderr,"\t%s - data len: %d format: %s\n", oc->option->name, data.len, oc->option->format );*/

	    /* first calculate length required */
	    dp = &(data.data[0]);
	    de = &(data.data[data.len]);
	    dbg_length = length;
	    dbg_d_last = dp;
	    for (fi=0;
		(oc -> option -> format [fi])
	       &&( (dp < de) 
		 ||( (dp == de) 
		   &&( (oc -> option -> format [fi] == 'A')
		     ||(oc -> option -> format [fi] == 'a')
		     )
	           )
		 );
		 fi++
		) 
	    {
/*		fprintf(stderr,"\t%d: %c\tdp: %p (%d) length: %d (%d)\n", 
			fi,  oc -> option -> format [fi],
			dp, dp - dbg_d_last, length, length - dbg_length  );
*/
		dbg_d_last = dp;
		dbg_length = length;
		switch ( oc-> option -> format [fi] )
		{
		case DHC_T_IP_ADDRESS:
		case DHC_T_UINT32:
		case DHC_T_INT32:
		    length += sizeof(uint32_t) + DHC_PAD( uint32_t, length );
		    dp += 4;
		    n_members++;
		    break;

		case DHC_T_UINT16:
		case DHC_T_INT16:
		    length += sizeof(int16_t) + DHC_PAD( int16_t, length );
		    dp += 2;
		    n_members++;
		    break;

		case DHC_T_ENUMERATION: 
		    for(i=0; oc->option->format[fi] && (oc->option->format[fi] != '.'); fi++, i++ );
		    /* the width of the enumeration data is meant to be specifiable, but currently is not -
                     * All enumeration values are currently bytes (uint8_t) - fall through
		     */
		case DHC_T_CHAR:
		case DHC_T_UCHAR:
		case DHC_T_BOOL:
		case DHC_T_IMPLICIT_BOOL:
		    length += sizeof(uint8_t) + DHC_PAD(uint8_t, length );
		    dp += 1;
		    n_members++;
		    break;

		case DHC_T_ENCAPSULATION:
		case DHC_T_ENCAPSULATION_NO_DIRECTIVE:
		    /* skip encapsulation name */
		    for(i=0; oc -> option -> format[fi] && (oc -> option -> format[fi] != '.'); fi++, i++);
		    /* treat encapsulations just like hex strings for now - 
                     * fall through 
		     */

		case DHC_T_HEX_STRING:
		    length += ( (unsigned long)de - (unsigned long)dp  );
		    dp = de ;
		    n_members++;
		    /* Hex strings will be exactly the remainder of the length,
		       and MAY not have a trailing nul byte.
		    */
		    break;

		case DHC_T_DOMAIN_NAME:
		case DHC_T_TEXT:
		case DHC_T_OPTION_SPACE:
		    length += ( data.len - length );
		    if( data.data[data.len - 1] != '\0' )
		    {
			length += 1;
		        /* text strings will always have at least one nul byte appended */
		    }
		    dp = de ;
		    n_members++;
		    break;
		    /* note: 
		       1. Arrays of strings are not allowed.
                       2. There can be only one string per struct
		       3. A string member must be the LAST member in a struct
                       So the above works OK.
		    */

		case DHC_T_ARRAY:
		case DHC_T_LIST:
		    is_array = 1;
		    member_size = length;
/*
		    fprintf(stderr, "\t\tn_elements: %lu member size: %d\n",
			    (((unsigned long)data.len) / ((unsigned long)(dp - &(data.data[0])))),
			    member_size
			   );			       
*/
		    n_elements = ( (unsigned long)data.len / (unsigned long )(dp - &(data.data[0])));
		    if( n_elements && n_members )
		    {
			switch( oc -> option -> format [ 0 ] )
			{
			case DHC_T_IP_ADDRESS:
			case DHC_T_UINT32:
			case DHC_T_INT32:
			    length += (element_pad = DHC_PAD(uint32_t, length));
			    break;

			case DHC_T_UINT16:
			case DHC_T_INT16:
			    length += (element_pad = DHC_PAD(uint8_t, length));
			    break;

			case DHC_T_CHAR:
			case DHC_T_UCHAR:
			case DHC_T_BOOL:
			case DHC_T_IMPLICIT_BOOL:
			case DHC_T_ENUMERATION: 
 
			    length += (element_pad = DHC_PAD(uint8_t, length));
			default:
			    break;
			}
		    }
		    length = n_elements * length;
		    length -= element_pad; /* no element after last! */
		    dp = de;
		    if( oc->option->format[fi+1] != '\0' )
		    {
			/* not allowed and would make us fail */
/*			fprintf(stderr, "Array list not the last in format - help!\n");*/
			goto dhcp4_client_option_add_fail ;
		    }
		    break;

		case DHC_T_OPTIONAL:
		case DHC_T_NONE:
		default:
		    break;		    
		}
	    }

/*	    fprintf(stderr,"\t%s - length: %d\n", oc->option->name, length);*/

	    if( length == 0 )
		return;

	    n_values = n_members * (n_elements ? n_elements : 1);

	    total_length =
		sizeof( DHCPv4_option ) 
	      + length  
	      + (  ( n_values > 1 )
		 ? ( ( n_values * sizeof(uint8_t*) )
		    +( DHC_PAD ( void*, length ) )
		   )
		 : 0
		);

/*	    fprintf(stderr,"\t%s - total_length: %d is_array: %d n_members: %d n_elements: %d member_size: %d n_values: %d\n", 
		    oc->option->name, total_length, is_array,    n_members,    n_elements,    member_size,    n_values);
*/
		    
	    option = malloc( total_length );	        

	    if( option == 0L )
		return ;	    
	    memset(option, '\0', total_length);
	    option->lease = lease;
	    option->name = strdup(oc->option->name);
	    option->format = strdup(oc->option->format);
	    option->code = oc->option->code;
	    option->unicode = u->index;
	    option->universe = strdup(u->name);
	    option->length = length;
	    option->form = 
		is_array
		? DHCO_ARRAY
		:( ( n_members > 1 )
		   ? DHCO_STRUCT
		   : DHCO_SCALAR 
		 );
	    if ( option->form == DHCO_STRUCT )
	    {
		option->size = length ;
		option->n_members = n_members;
		option->n_elements = 0;
	    }else 
	    if ( option->form == DHCO_ARRAY )
	    {
		option->size = member_size;
		option->n_elements = n_elements;
		option->n_members = n_members;
	    }

/*	    fprintf(stderr, "\tVL size: %d address: %p offset: %d\n",
		    ( n_values * sizeof(uint8_t*) ),
		    ( n_values > 1 ) ?
		    ( (unsigned long)&(option->value[length]) 
		    + (unsigned long)(DHC_PAD(void*,length))
		    ) : 0, DHC_PAD(void*,length)
		);
*/
	    vl = (void*)
		 ( (option->form != DHCO_SCALAR) 
		   ?( (unsigned long)&(option->value[length]) 
		    + (unsigned long)(DHC_PAD(void*,length))
		    )
		   : 0L
		 );
	        
/*	    fprintf(stderr,"\toption:\n\tname: %s\n\tformat:%s\n\tuniverse: %d\n\tcode: %d\n\tform: %d\n\tn_members:%d\n\tn_elements: %d\n\tlength: %d\n\t&value[0]: %p\n\tvl: %p\n\tde: %p\n",
		    option->name, option->format, option->universe, option->code, option->form, 
		    option->n_members, option->n_elements, option->length, &(option->value[0]), vl, 
		    &(data.data[data.len]) );
*/
	    
	    value = &(option->value[0]);
	    endv = &(option -> value[ length ]);	    
	    dp = &(data.data[0]);
	    de = &(data.data[data.len]);	    
	    length = 0;
	    dbg_length = length;
	    dbg_d_last = dp;
	    dbg_v_last = value;
	    dbg_vl_last = vl;

	    for ( fi=0;  
		  (option -> format [fi]) 
	       && (dp < de)
	       && (value < endv);
		  fi++
		) 
	    {
		if ( vl )
		    *vl = value;

/*		fprintf(stderr, "\t%d %c\tdp:%p (%d) value: %p (%d) vl: %p (%lu) length: %d(%d)\n",
			fi, oc -> option -> format [fi], dp, dp - dbg_d_last,
			value, value - dbg_v_last,
			vl, (unsigned long)vl - (unsigned long)dbg_vl_last,
			length, (length - dbg_length)
		    );
*/
		dbg_d_last = dp;
		dbg_v_last = value;
		dbg_vl_last = vl;
		dbg_length = length;

		switch ( oc -> option -> format [fi] )
		{
		case DHC_T_IP_ADDRESS:
		    value += DHC_PAD(uint32_t, length);
		    length += DHC_PAD(uint32_t, length);
		    memcpy(value, dp, sizeof(uint32_t));
		    value  += sizeof(uint32_t);
		    length += sizeof(uint32_t);
		    dp += 4;
		    break;

		case DHC_T_INT32:
		case DHC_T_UINT32:
		    value += DHC_PAD(uint32_t, length);
		    length += DHC_PAD(uint32_t, length);
		    memcpy( &intv, dp, sizeof(uint32_t));
		    *((uint32_t*)value) = ntohl( intv );
		    value += sizeof(uint32_t);
		    length += sizeof(uint32_t);
		    dp += 4;
		    break;

		case DHC_T_UINT16:
		case DHC_T_INT16:
		    value += DHC_PAD(uint16_t, length);
		    length += DHC_PAD(uint16_t, length);
		    memcpy( &sv, dp, sizeof(uint16_t));
		    *((uint16_t*)value)=ntohs( sv );
		    value += sizeof(uint16_t);
		    length += sizeof(uint16_t);
		    dp += 2;
		    break;

		case DHC_T_ENUMERATION: 
		    for(i=0; oc->option->format[fi] && (oc->option->format[fi] != '.'); fi++, i++ );
		    /* the width of the enumeration data is meant to be specifiable, but currently is not -
                     * All enumeration values are currently bytes (uint8_t) - fall through
		     */
		case DHC_T_CHAR:		    
		case DHC_T_UCHAR:
		case DHC_T_BOOL:
		case DHC_T_IMPLICIT_BOOL:
		    value += DHC_PAD(uint8_t, length );
		    length += DHC_PAD(uint8_t, length );
		    *value = *dp;
		    value  += sizeof(uint8_t);
		    length += sizeof(uint8_t);
		    dp += 1;
		    break;

		case DHC_T_ENCAPSULATION:
		case DHC_T_ENCAPSULATION_NO_DIRECTIVE:
		    /* skip encapsulation name */
		    for(i=0; oc -> option -> format[fi] && (oc -> option -> format[fi] != '.'); fi++, i++);
		    /* treat encapsulations just like hex strings for now - 
                     * fall through 
		     */
		case DHC_T_HEX_STRING:
		    memcpy( value, dp, ((unsigned long)de) - ((unsigned long)dp) );
		    value += ((unsigned long)de) - ((unsigned long)dp);
		    length += ((unsigned long)de) - ((unsigned long)dp);
		    dp = de;
		    break;

		case DHC_T_DOMAIN_NAME:
		case DHC_T_TEXT:
		case DHC_T_OPTION_SPACE:
		    p = value;
		    memcpy( value, dp, ((unsigned long)de) - ((unsigned long)dp) );		    
		    value += ((unsigned long)de) - ((unsigned long)dp);
		    if( *(de-1) != '\0' )
		    {
			*(value++) = '\0';
			length += 1;
		    }
		    length += (((unsigned long)value) - ((unsigned long)p));
		    dp = de;
		    break;
				    
		case DHC_T_ARRAY:
		case DHC_T_LIST:
		    fi = -1;
		    continue;

		case DHC_T_OPTIONAL:
		case DHC_T_NONE:
		    /*XXX to be done!*/
		    break;
		}
		if ( vl )
		    vl++;
	    }

/*	    fprintf(stderr," final length: %d option length: %d\n", length, option->length );*/

	    if ( (dp != de) || (value != endv) )
	    {
/*		fprintf(stderr,"fail: dp: %p de: %p value: %p endv: %p\n",
			dp, de, value, endv
		    );
*/
		goto dhcp4_client_option_add_fail;
	    }
	    
	    dbg_t = tsearch( option,  &(lease->options), option_code_comparator );
	    dbg_t = tsearch( option,  &(lease->options_by_name), option_name_comparator );
	    
/*	    fprintf(stderr, "\ttsearch: %p\n", dbg_t);*/
	    return;

	dhcp4_client_option_add_fail:
/*	    fprintf(stderr,"dhcp4_client_option_add FAILED!\n");*/
	    return;
	}
    }
}

struct dhcpv4_pack_s
{
    DHCPv4_lease *lease;
    uint8_t *buf;
    uint8_t *bufp;
    uint32_t buflen;
};

void dhcpv4_pack_lease_option(DHCPv4_option *opt, void *psp )
{
    struct dhcpv4_pack_s *ps = psp;
    uint32_t n_values =
	( (opt->n_elements ? opt->n_elements : 1)
	 * opt->n_members
        ),
	name_len = strlen(opt->name),
	uni_len = strlen(opt->universe),
	fmt_len = strlen(opt->format);
/*        uint8_t *dbg_bufp = ps->bufp;*/

    if( ps->buf )
	*((uint32_t*)(ps->bufp))=name_len;
    ps->bufp +=sizeof(uint32_t);
    if( ps->buf )
	memcpy(ps->bufp,opt->name,name_len);
    ps->bufp+=name_len + DHC_PAD(uint32_t,name_len);
       
    if( ps->buf )
	*((uint32_t*)(ps->bufp))=uni_len;
    ps->bufp+=sizeof(uint32_t);
    if( ps->buf )
	memcpy(ps->bufp,opt->universe,uni_len);
    ps->bufp+=uni_len +  DHC_PAD(uint32_t,uni_len);

    if( ps->buf )
	*((uint32_t*)(ps->bufp))=fmt_len;
    ps->bufp+=sizeof(uint32_t);
    if( ps->buf )
	memcpy(ps->bufp,opt->format,fmt_len);
    ps->bufp+=fmt_len +  DHC_PAD(uint32_t,fmt_len);

    uint8_t sizeof_opt = ((unsigned long)&(opt->value[0]))-((unsigned long)opt);
    if( ps->buf )
	memcpy(ps->bufp, opt, sizeof_opt);
    ps->bufp += sizeof_opt;
    if( ps->buf )
	memcpy(ps->bufp, &(opt->value[0]), opt->length);
    ps->bufp +=  opt->length
	      +  DHC_PAD(uint32_t, opt->length)
	      +  ( ((n_values > 1) ? n_values : 0)  * sizeof(void*));

/*    fprintf (stderr,"opt: %s %d (%d %d) %d %d\n",opt->name, ps->bufp - dbg_bufp, opt->length, DHC_PAD(uint32_t,opt->length), n_values,  ps->bufp - ps->buf  );*/
}

int dhcpv4_pack_lease( DHCPv4_lease *lease, uint8_t *buf, uint32_t len )
{
    if(lease == 0L)
	return 0;

    int filename_len   = lease->filename ? strlen(lease->filename) : 0,
	servername_len = lease->server_name ? strlen(lease->server_name) : 0,
	ifname_len = lease->if_name ? strlen(lease->if_name) : 0;

    struct dhcpv4_pack_s ps =
    {   .lease = lease,
        .buf   = buf,
	.bufp  = buf,
        .buflen= len
    };
        
    if(ps.buf)
    {
	*((uint32_t*)(ps.bufp)) = len;
	ps.bufp += sizeof(uint32_t);
    }else
	ps.bufp += sizeof(uint32_t);

    if( ps.buf )
	memcpy(ps.bufp, lease, sizeof(DHCPv4_lease));

    ps.bufp += sizeof(DHCPv4_lease) + DHC_PAD(uint32_t, sizeof(DHCPv4_lease));   

    if( filename_len )
    {
	if( ps.buf )
	    *((uint32_t*)ps.bufp) = filename_len;
	ps.bufp+=sizeof(uint32_t);
	if( ps.buf )
	    memcpy(ps.bufp, lease->filename, filename_len);
	ps.bufp += filename_len;
	ps.bufp += DHC_PAD(uint32_t, filename_len);
    }else
    {
	if( ps.buf )
	    *((uint32_t*)ps.bufp) = 0;
	ps.bufp+=sizeof(uint32_t);
    }

    if( servername_len )
    {
	if( ps.buf )
	    *((uint32_t*)ps.bufp) = servername_len;
	ps.bufp+=sizeof(uint32_t);
	if( ps.buf )
	    memcpy(ps.bufp, lease->server_name, servername_len);
	ps.bufp += servername_len;
	ps.bufp += DHC_PAD(uint32_t, servername_len);
    }else
    {
	if( ps.buf )
	    *((uint32_t*)ps.bufp) = 0;
	ps.bufp+=sizeof(uint32_t);
    }

    if( ifname_len )
    {
	if( ps.buf )
	    *((uint32_t*)ps.bufp) = ifname_len;
	ps.bufp+=sizeof(uint32_t);
	if( ps.buf )
	    memcpy(ps.bufp, lease->if_name, ifname_len);
	ps.bufp += ifname_len;
	ps.bufp += DHC_PAD(uint32_t, ifname_len);	
    }else
    {
	if( ps.buf )
	    *((uint32_t*)ps.bufp) = 0;
	ps.bufp+=sizeof(uint32_t);
    }	

    dhcpv4_process_options( lease, dhcpv4_pack_lease_option, &ps );
        
    return (ps.bufp - ps.buf);
}

DHCPv4_lease *dhcpv4_unpack_lease( uint8_t *buf )
{
    if( buf == 0 )
	return 0;

    uint32_t packlen = *((uint32_t*)buf);
    uint8_t *bufp = buf + sizeof(uint32_t) /*, *dbg_bufp=bufp*/;    
    DHCPv4_lease  *rlease = 0;
    DHCPv4_option *opt=0, *ropt=0;
    uint32_t n_values=0, slen=0;

    if( packlen < (sizeof(DHCPv4_lease) + (2 * sizeof(uint32_t))) )
	return 0L;

    rlease = calloc(1, sizeof(DHCPv4_lease));    

    memcpy(rlease, bufp, sizeof(DHCPv4_lease));
    rlease->options = 0;
    rlease->options_by_name = 0;
    rlease->handler = 0;
    rlease->handler_arg = 0;
    rlease->if_name = 0;
    rlease->server_name = 0;
    rlease->filename = 0;

    bufp += sizeof(DHCPv4_lease) + DHC_PAD(uint32_t, sizeof(DHCPv4_lease));
    
    if ((slen = *((uint32_t*)bufp)) > 0)
    {	
	rlease->filename = calloc(1, slen + 1 );
	bufp += sizeof(uint32_t);
	memcpy(rlease->filename, bufp, slen); 
	bufp += slen + DHC_PAD(uint32_t, slen);
    }else
	bufp += sizeof(uint32_t);
    
    if((slen = *((uint32_t*)bufp)) > 0)
    {
	rlease->server_name = calloc(1, slen + 1 );
	bufp += sizeof(uint32_t);
	memcpy(rlease->server_name, bufp, slen); 
	bufp += slen + DHC_PAD(uint32_t, slen);
    }else
	bufp += sizeof(uint32_t);    
 
    if((slen = *((uint32_t*)bufp)) > 0)
    {
	rlease->if_name = calloc(1, slen + 1 );
	bufp += sizeof(uint32_t);
	memcpy(rlease->if_name, bufp, slen); 
	bufp += slen + DHC_PAD(uint32_t, slen);
    }else
	bufp += sizeof(uint32_t);        

    while( bufp < (buf + packlen) )
    {
	char *name, *universe, *format;
	void **lv, *lv0,  **rlv;
	int nv=0;

	/* dbg_bufp = bufp; */

	slen = *((uint32_t*)bufp);
	name = calloc(1, slen + 1);
	bufp += sizeof(uint32_t);
	memcpy(name, bufp, slen);

	bufp += slen + DHC_PAD(uint32_t, slen);

	slen = *((uint32_t*)bufp);
	universe = calloc(1, slen + 1);
	bufp += sizeof(uint32_t);
	memcpy(universe, bufp, slen);
	bufp += slen + DHC_PAD(uint32_t, slen);

	slen = *((uint32_t*)bufp);
	format = calloc(1, slen + 1);
	bufp += sizeof(uint32_t);
	memcpy(format, bufp, slen);
	bufp += slen + DHC_PAD(uint32_t, slen);
	
	opt = (DHCPv4_option*)bufp;

	uint8_t sizeof_opt = ((unsigned long)&(opt->value[0]))-((unsigned long)opt);

	n_values = 
	    ( (opt->n_elements ? opt->n_elements : 1)
	     * opt->n_members
	    );

/*	fprintf(stderr, "\t%s %d %d %d\n", name, n_values, opt->n_elements, opt->n_members);*/
		
	ropt = 
	    calloc
	    (1,  sizeof_opt
	      +  opt->length 
	      +  DHC_PAD(uint32_t, opt->length)
	      +  ( ((n_values > 1) ? n_values : 0)  * sizeof(void*))
	    );
	memcpy(ropt, opt, 
	         sizeof_opt
	      +  opt->length 
	      +  DHC_PAD(uint32_t, opt->length)
	      +  ( ((n_values > 1) ? n_values : 0)  * sizeof(void*))
	      );
	ropt->name = name;
	ropt->universe = universe;
	ropt->format = format;
	
	if( n_values > 1 )
	{
	    lv =  (void*) 
		(  &(opt->value[0])
		+  opt->length 
		+  DHC_PAD(uint32_t, opt->length)
		);

	    rlv = (void*)
		(  &(ropt->value[0])
		+  opt->length 
		+  DHC_PAD(uint32_t, opt->length)
		);
	    
	    lv0 = *lv;
	    nv = n_values;
	    while(nv--)
		*(rlv++)=&(ropt->value[0]) + ((*(lv++))-lv0);
	}
	bufp +=  sizeof_opt
	      +  opt->length 
	      +  DHC_PAD(uint32_t, opt->length)
	      +  ( ((n_values > 1) ? n_values : 0)  * sizeof(void*))
	    ;
/*	fprintf(stderr,"unpack: %s %d %d\n", name , bufp - dbg_bufp, n_values); */
	ropt->lease = rlease;
	tsearch(ropt,&(rlease->options),option_code_comparator);
	tsearch(ropt,&(rlease->options_by_name),option_name_comparator);
    }
    return rlease;
}
