/* dhcp4_nic.c
 *
 * Network Interface Configurator for the ISC DHCP IPv4 client library.
 *
 *  Copyright(C) Jason Vas Dias <jvdias@redhat.com> Red Hat Inc. May 2006
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation at 
 *           http://www.fsf.org/licensing/licenses/gpl.txt
 *  and included in this software distribution as the "LICENSE" file.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */
#include <sys/types.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <dhcp4_nic.h>
#include <dhcp4client.h>
#include <isc_dhcp/dhcpd.h>

struct dhcpv4_control_s
{
/*{ this structure MUST be the same as in libdhcp_control.h for DHCP clients */
    LIBDHCP_Callback    callback;    	/* the DHCP clients' main loop calls this on state changes */
    uint16_t            capability;     /* LIBDHCP_Capability bits to enable                       */
    uint8_t             finished;       /* set to one to make clients exit their main loop         */
    uint8_t             decline;        /* set to one to decline the lease (DHCPv4 only)           */
    time_t              timeout;        /* (timeout+now) == time after which clients MUST return   */
    time_t              now;            /* clients set this to time(0) on entering main loop       */
    void               *arg;            /* user data pointer                                       */
    LIBDHCP_Error_Handler eh;
/*} end of fields known to DHCP clients */
    uint8_t             log_level;      /* maximum log level (LOG_FATAL excluded)                  */    
/* struct DHCPv4_nic { */
    NLH_t               nh;
    NIC_t               nic;
    IPaddr_list_t       addr_list;
    IProute_list_t      route_list;
    IPaddr_list_t       dns_list;
    char               *search_list;
    char               *host_name;
    DHCPv4_lease       *lease;
/*                   } */
    IPaddr_t           ip4;
    uint32_t           mtu;
    char               *if_name;
    char              **argv;
    int                 argc;
};

static void
dhc4_log( DHCPv4_control *ctl, int priority, char *fmt, ...)
{
    va_list va;
    if((ctl == 0)|| ((ctl->eh == 0) || (ctl->log_level < priority)))
	return;
    va_start(va, fmt);
    ctl->eh((LIBDHCP_Control*)ctl, priority, fmt, va);
    va_end(va);
}

DHCPv4_control*
dhcpv4_control_va
( 
    NLH_t                  nh,
    char                  *eth_if_name,
    LIBDHCP_Capability     dhc_cap, 
    time_t                 timeout,
    LIBDHCP_Error_Handler  error_handler,
    uint8_t                log_level,
    va_list                va
)
{
    DHCPv4_control *dhc = calloc( 1, sizeof( DHCPv4_control ) ); 
    char **argv, *a[32]={strdup("dhclient")}, *p;
    int argc,i;
    uint8_t d_needed=1, if_needed=1;

    if( dhc == 0L )
	return 0;

    dhc->nh = nh;

    if( (dhc->nic = nic_by_name(dhc->nh, eth_if_name)) == 0L ) {
	free(dhc);
	return 0;
    }

    dhc->capability = dhc_cap;
    dhc->callback = dhcp4_nic_callback;
    dhc->timeout = timeout;
    dhc->arg = 0;
    dhc->eh = error_handler;
    dhc->log_level = log_level;
    dhc->if_name = strdup(eth_if_name);
    
    for(argc=1, argv=&(a[1]); (argc < 31) && (p = va_arg(va, char *)); argc++, argv++)
    {
	*argv = strdup(p);
	if( strcmp(p,"-d") == 0)
	    d_needed = 0;
	if( strcmp(p,eth_if_name) == 0)
	    if_needed = 0;	
    }

    if ( d_needed )
    {
	if ( argc > 31 )
	{
	    dhcpv4_control_free(dhc);
	    return 0L;
	}
	a[argc++] = strdup("-d");
    }

    if ( if_needed )
    {
	if ( argc > 31 )
	{
	    dhcpv4_control_free(dhc);
	    return 0L;
	}
	a[argc++] = strdup(eth_if_name);
    }

    a[argc] = 0L;
    argv = calloc(argc+1, sizeof(char*));
    if( argv == 0 )
    {
	dhcpv4_control_free(dhc);
	return 0L;
    }
    dhc->argc = argc;    
    for(i=0; i < argc; i++)
	argv[i] = a[i];
    argv[i]=0;
    dhc->argv = argv;
    STAILQ_INIT( &(dhc->addr_list) );
    STAILQ_INIT( &(dhc->route_list) );
    STAILQ_INIT( &(dhc->dns_list) );    
    return dhc;
}

DHCPv4_control*
dhcpv4_control
( 
    NLH_t                  nh,
    char                  *eth_if_name,
    LIBDHCP_Capability     dhc_cap, 
    time_t                 timeout,
    LIBDHCP_Error_Handler  error_handler,
    uint8_t                log_level,
    ...
)
{
    va_list va;
    va_start( va, log_level );
    DHCPv4_control * dhc =
    dhcpv4_control_va
    ( 
	nh,
	eth_if_name,
	dhc_cap, 
        timeout,
	error_handler,
	log_level,
	va
    );   
    va_end(va);
    return(dhc);
}

void dhcpv4_control_free( DHCPv4_control *dhc )
{
    if(dhc->lease)
    {
	dhcpv4_lease_free(dhc->lease);
	dhc->lease = 0;
    }
    if(dhc->if_name)
    {
	free(dhc->if_name);
	dhc->if_name = 0;
    }
    if( dhc->argv )
    {
	char **p;
	for(p=dhc->argv; *p; p++)
	    free(*p);
	free(dhc->argv);
	dhc->argv = 0;
    }
    if( !STAILQ_EMPTY( &(dhc->addr_list) ) )
	nic_address_list_free(&(dhc->addr_list));
    if( !STAILQ_EMPTY( &(dhc->dns_list) ) )
	nic_address_list_free(&(dhc->dns_list));
    if( !STAILQ_EMPTY( &(dhc->route_list) ) )
	nic_route_list_free(&(dhc->route_list));
    
    if( dhc->search_list )
	free(dhc->search_list);
    if( dhc->host_name )
	free(dhc->host_name);

    free(dhc);
}

extern char **environ;

DHCPv4_nic *do_dhcpv4( DHCPv4_control *dh4c )
{
    dhcpv4_client((LIBDHCP_Control*)dh4c, dh4c->argc, dh4c->argv, environ);

    if(dh4c->lease == 0)
    { /* empty lease / unable to contact server */
	dhcpv4_control_free( dh4c );
	return 0;
    }
    return (DHCPv4_nic*)&(dh4c->nh);
}

NIC_Res_t dhcpv4_nic(NLH_t nh, DHCPv4_nic *nic )
{
    return 
	nic_configure
        (
	    nh,
	    nic->nic,
	    &(nic->address_list),
	    &(nic->route_list),
	    &(nic->dns_list),
	    nic->search_list,
	    nic->host_name
	);
}

DHCPv4_nic *dhcp4_set_lease(DHCPv4_control *ctl, DHCPv4_lease *lease)
{
    ctl->lease = lease;
    return (DHCPv4_nic*)&(ctl->nh);
}

int dhcp4_nic_callback
(   LIBDHCP_Control *cp,
    DHCP_State state, 
    void *arg
)
{
    DHCPv4_control *control = (DHCPv4_control *) cp;
    char buf[32];
    dhc4_log
    (   control, LOG_DEBUG,
	"DHCPv4 %s  - state: %s",
	control->if_name,
	libdhcp_state_string(state,&(buf[0]))
    );
    switch( state )
    {
    case DHC_TIMEDOUT:
	dhc4_log
	(   control, LOG_INFO,
	    "DHCPv4 %s  - TIMED OUT.",
	    control->if_name,
	    libdhcp_state_string(state,&(buf[0]))
	);
	control -> finished = 1;
	break;
	
    case DHC4_PREINIT:
    {	
    }   break;

    case DHC4_BOUND:
    case DHC4_REBOOT:
    case DHC4_RENEW:
    case DHC4_REBIND:
    case DHC4_TIMEOUT: /* Cannot currently contact server, but we have an active lease */
    {	 
	control -> lease = dhcpv4_lease( arg );
	dhc4_log
	(   control, LOG_DEBUG,
	    "DHCPv4 %s - BOUND: %p", control->if_name, dhcpv4_lease
	);
	control -> finished = 1;
	break;
    }

    case DHC4_RELEASE:
    case DHC4_EXPIRE:
    case DHC4_FAIL:
    case DHC4_STOP:
	/* unconfigure existing lease */
	/* XXX - should do something here!*/
	control -> finished = 1;
	break;

    default:
	dhc4_log
	(   control, LOG_ERR,
	    "DHCPv4 %s - entered unhandled state.",
	    control->if_name
	);
    }
    return 0;
}

int dhcp4_process_lease(DHCPv4_control *ctl)
{
    if( ( ctl == 0L ) || (ctl->lease == 0L) ) return(0);
    /* process new lease */
    DHCPv4_lease *lease = ctl->lease;
    char buf[32];
    ip_addr_t lease_address;
        
    ctl->lease = lease;
    
    lease_address = ip_addr_in( &lease->address );
    
    dhc4_log
    (   ctl, LOG_INFO,
	"DHCPv4 %s - obtained lease %s",
	ctl->if_name,
	ip_text_addr(&lease_address, buf, 32)
    );
    ctl->lease = lease;
    ctl->ip4 = nic_addr(ctl->nh, lease_address );
    
    IPaddr_list_node_t *n = calloc(1,sizeof(IPaddr_list_node_t));
    n->addr = ctl->ip4;
    STAILQ_INSERT_TAIL( &(ctl->addr_list), n, link );

    dhcpv4_process_options(lease, dhcp4_nic_option_handler, ctl);

    return ((!STAILQ_EMPTY(&(ctl->addr_list))) || (ctl->lease->options != 0));
}

uint32_t dhcpv4_mtu_option( DHCPv4_control *ctl )
{
    return ctl->mtu;
}

void dhcp4_nic_option_handler( DHCPv4_option *option, void *arg )
{
    DHCPv4_control *control = arg;
    char buf[32];
    
    if ( option -> unicode == DHC_DHCP_Universe )
    {
	switch( option -> code )
	{
	case DHCO_SUBNET_MASK:
	{
	    dhc4_log
	    (   control, LOG_INFO, "DHCPv4 %s - option subnet-mask: %s", 
		control->if_name,
		inet_ntop( AF_INET, (struct in_addr*)&(option->value[0]), &(buf[0]), sizeof(buf))
	    );
	    ip_addr_t ip4 = nic_ip_addr(control->ip4);
	    uint32_t sm=0;
	    uint32_t  i=32, b=1;
	    memcpy(&sm, &(option->value[0]), sizeof(uint32_t));
	    sm = ntohl(sm);
	    for(; i && ((sm & b) != b); i-=1, b <<= 1);
	    nic_addr_set_prefix( control->ip4, i);
	    ip_addr_t ip4_broadcast = ip_v4_broadcast( &ip4, i );
	    nic_addr_set_broadcast( control->ip4, ip4_broadcast );
	    dhc4_log
	    (   control, LOG_INFO, "DHCPv4 %s - option subnet-mask - prefix_len: %d broadcast: %s", 
		control->if_name, i, ip_text_addr(&ip4_broadcast, buf, 32)		
	    );
	}
	break;

	case DHCO_ROUTERS:
	{
	    int i;
	    struct in_addr *routers = (struct in_addr*) &(option->value[0]);
	    dhc4_log
	    (   control, LOG_INFO, "DHCPv4 %s - option routers:", 
		control->if_name
	    );
	
	    for( i = 0; i < option->n_elements; i++, routers++ )
	    {
		ip_addr_t gw = ip_addr_in( routers );

		dhc4_log
	        (   control, LOG_DEBUG, "DHCPv4 %s - option routers default gateway %d: %s", 
		    control->if_name, i,
		    ip_text_addr(&gw,&(buf[0]),32)
		);

		IProute_t route =
		    nic_route_new
		    (   control->nh,
			nic_get_index(control->nic),
			0L,0,                      /* DEFAULT ROUTE ! */
			&gw,                       /* gateway */
			-1,                        /* -1: default: RTN_UNICAST */
            -1,                        /* -1: default: RTPROT_BOOT */
			-1,                        /* default scope: UNIVERSE */
			(i > 0) ? (int8_t)i : -1,  /* priority: inc if more than one */
			-1,                        /* table: default (local) */
			0L/*no iif*/, 0L, 0/*no src*/		    
		    );		
		IProute_list_node_t *n = calloc(1,sizeof(IProute_list_node_t));
		n->route = route;
		STAILQ_INSERT_TAIL(&(control->route_list),n,link);
	    }
	}
	break;	
	case DHCO_STATIC_ROUTES:
	{
	    dhc4_log
	    (   control, LOG_DEBUG, "DHCPv4 %s - option static-routes:", 
		control->if_name
	    );

	    char buf2[32];
	    int i;
	    struct { struct in_addr dst, gw; } *routes = (void*)&(option->value[0]);
	    IProute_t route;

	    for( i = 0; i < option->n_elements; i++, routes++ )
	    {
		ip_addr_t 
		    dst=ip_addr_in( &(routes->dst) ),
		    gw=ip_addr_in( &(routes->gw) );
		    
		dhc4_log
		(   control, LOG_DEBUG, "DHCPv4 %s - option static-routes - route %d: %s via %s",
		    control->if_name, 
		    ip_text_addr( &dst, &(buf[0]), sizeof(buf)),
		    ip_text_addr( &gw , &(buf2[0]),sizeof(buf))
		);

		uint32_t dip= ip_v4_addr(&dst);
		uint32_t  dst_len=32, b=1;
		for(; dst_len && ((dip & b) != b); dst_len--, b <<= 1);
		route = 
		    nic_route_new
		    (   control->nh,
			nic_get_index(control->nic),
			&dst, dst_len,
			&gw,                       /* gateway */
			-1,                        /* -1: default: RTN_UNICAST */
                        -1,                        /* -1: default: RTPROT_BOOT */
			-1,                        /* default scope: global */
			(i > 0) ? (int8_t)i : -1,  /* priority: increasing if more than one */
			-1,                        /* table: default (local) */
			0L/*no iif*/, 0L,0/*no src*/		    
		    );			       
		IProute_list_node_t *n = calloc(1,sizeof(IProute_list_node_t));
		n->route = route;
		STAILQ_INSERT_TAIL(&(control->route_list),n,link);
	    }
	}
	break;

	case DHCO_DOMAIN_NAME:
	    dhc4_log
	    (   control, LOG_DEBUG, "DHCPv4 %s - option domain-name: %s",
		control->if_name, (char*)(option->value)
	    );
	    control->search_list = strdup( (char*)option->value );
	    break;

	case DHCO_HOST_NAME:
	    dhc4_log
	    (   control, LOG_DEBUG, "DHCPv4 %s - option host-name: %s",
		control->if_name, (char*)(option->value)
	    );
	    control->host_name = strdup( (char*)option->value );
	    break;

	case DHCO_DOMAIN_NAME_SERVERS:
	{
	    dhc4_log
	    (   control, LOG_DEBUG, "DHCPv4 %s - option domain-name-servers:",
		control->if_name
	    );
	    int i;
	    struct in_addr *dns = (struct in_addr*) &(option->value[0]);
	    for( i = 0; i < option->n_elements; i++, dns++ )
	    {
		dhc4_log
	        (   control, LOG_DEBUG, "DHCPv4 %s - domain-name-server %d: %s",
		    control->if_name, i,
		    inet_ntop(AF_INET, dns, buf, sizeof(buf))
		);
		IPaddr_t dnsIP = nic_addr(control->nh, ip_addr_in(dns));
		IPaddr_list_node_t *n = calloc(1,sizeof(IPaddr_list_node_t));
		n->addr = dnsIP;
		STAILQ_INSERT_TAIL(&(control->dns_list), n, link);
	    }
	}
	break;
	case DHCO_INTERFACE_MTU:
	{
	    dhc4_log
	    (   control, LOG_DEBUG, "DHCPv4 %s - option interface-mtu: %d",
		control->if_name, *((uint32_t*)&(option->value[0]))
            );
	    control->mtu =  *((uint32_t*)&(option->value[0]));
	}
	break;
	default:
	    /* XXX: many more options to do something about! */
	    break;
	}
    }
}
