/* dhcp6_lease.c
 *
 *  Representation of DHCPv6 lease options for libdhcp
 *
 * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of
 * the GNU General Public License v.2.  This program is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY expressed or
 * implied, including the implied warranties of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.  You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA. Any Red Hat trademarks that are incorporated in the source code or
 * documentation are not subject to the GNU General Public License and may
 * only be used or replicated with the express permission of Red Hat, Inc.
 *
 * Red Hat Author(s): Jason Vas Dias
 *                    David Cantrell
 */

#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <resolv.h>
#include <string.h>
#include <time.h>
#include <stdio.h>

/* dhcpv6 includes: */

#include <sys/queue.h>
/* #include <queue.h> 
 * dhcpv6 uses an older copy of BSD queue.h, which 
 * is happily compatible with the new version in glibc
 */
#define  LIBDHCP
#include <dhcp6.h>
#include <cfg.h>
#include <lease.h>
#undef malloc
#undef realloc
#undef calloc
#undef free
#undef strdup
#include <malloc.h>
#include <dhcp6_lease.h>


DHCPv6_lease *dhcpv6_lease( struct dhcp6_optinfo *optinfo )
{
    if ( TAILQ_EMPTY( &(client6_iaidaddr.lease_list) ) )
	return 0L;

    DHCPv6_lease *lease = calloc(1, sizeof(DHCPv6_lease));
				 
    if( lease == 0L )
	return 0;
    
    STAILQ_INIT(&(lease->addresses));
    STAILQ_INIT(&(lease->dns));

    struct dhcp6_lease *l;
    for (l = TAILQ_FIRST( &(client6_iaidaddr.lease_list) ); l; l = TAILQ_NEXT(l,link))
    {
	if(  ( l->state == DHCP6_ACTIVE )
	  && ( l->lease_addr.type != DHCPv6_DELEGATED_PREFIX )
	  && (  ( l->lease_addr.status_code  == DH6OPT_STCODE_SUCCESS ) 
	      ||( l->lease_addr.status_code  == DH6OPT_STCODE_UNDEFINE )
	     )
	  ) 
	{
	    DHCPv6_lease_address *addr = calloc(1, sizeof( DHCPv6_lease_address ) );
	    if( addr == 0 )
		goto dhcpv6_lease_failure;
	    
	    memcpy(&(addr->address), &(l->lease_addr.addr), sizeof(struct in6_addr));
	    
	    addr -> prefix_length =  l->lease_addr.plen;
	    addr -> type =  l->lease_addr.type;
	    addr -> status = l->lease_addr.status_code;
    	    addr -> time_acquired = l->start_date;
	    addr -> prefer_lifetime = l->lease_addr.preferlifetime;
	    addr -> valid_lifetime = l->lease_addr.validlifetime;
	    
	    STAILQ_INSERT_TAIL( &(lease->addresses), addr, link );
	}
    }

    lease->rebind_time = client6_iaidaddr.client6_info.iaidinfo.rebindtime;
    lease->renew_time =  client6_iaidaddr.client6_info.iaidinfo.renewtime;
    lease->iaid = client6_iaidaddr.client6_info.iaidinfo.iaid;
    lease->client_duid = malloc(client6_iaidaddr.client6_info.clientid.duid_len );
    if( lease->client_duid  == 0L )
	goto dhcpv6_lease_failure;
    memcpy(lease->client_duid,  client6_iaidaddr.client6_info.clientid.duid_id,
	                        client6_iaidaddr.client6_info.clientid.duid_len);
    lease->client_duid_len = client6_iaidaddr.client6_info.clientid.duid_len;
    lease->server_duid = malloc(client6_iaidaddr.client6_info.serverid.duid_len );
    lease->server_duid_len = client6_iaidaddr.client6_info.serverid.duid_len;
    if( lease->server_duid  == 0L )
	goto dhcpv6_lease_failure;
    memcpy(lease->server_duid,  client6_iaidaddr.client6_info.serverid.duid_id,
	                        client6_iaidaddr.client6_info.serverid.duid_len);    
    memcpy(&(lease->server_address), &(optinfo->server_addr), sizeof(struct in6_addr) );
    if( (lease->if_name = strdup( client6_iaidaddr.ifp->ifname )) == 0L )
 	goto dhcpv6_lease_failure;    
    lease->if_index = if_nametoindex( lease->if_name );
    
    if ( ! TAILQ_EMPTY( &(optinfo -> dns_list.addrlist) ) )
    {
	struct dhcp6_listval *v;	
	IPv6_address *dns;
	int i=0;
	for (v = TAILQ_FIRST(&(optinfo->dns_list.addrlist)); v && (i < MAXNS); v = TAILQ_NEXT(v, link), i++) 
	{
	    dns = calloc(1, sizeof(IPv6_address));
	    if( dns == 0L )
		goto dhcpv6_lease_failure;
	    memcpy(&(dns->address), &(v->val_addr6), sizeof(struct in6_addr));
	    STAILQ_INSERT_TAIL(&(lease->dns), dns, link);
	} 
    }
    if ( optinfo -> dns_list.domainlist != 0 )
    {
	int i=0, len;
	struct domain_list *dlist;
	char search_list[256], *p=search_list;
	for (dlist = optinfo->dns_list.domainlist; dlist && (i < MAXDNSRCH); dlist = dlist->next, i++)
	{
	    len = strlen(dlist->name);
	    if ( (len + 1 + (p - search_list)) >= sizeof(search_list) )
		break;
	    p += sprintf(p,"%s ", dlist->name);
	}
	if( p - search_list )
	    if( ( lease->search_list = strdup(search_list) ) == 0L )
		goto dhcpv6_lease_failure;
    }
    
    return lease;

    dhcpv6_lease_failure:
    dhcpv6_lease_free( lease );
    return 0L;
}

void dhcpv6_lease_free( DHCPv6_lease *lease )
{
    if( !STAILQ_EMPTY(&(lease->addresses)) )
    {
	DHCPv6_lease_address *addr = STAILQ_FIRST( &(lease->addresses) ), *next;
	do
	{
	    next = STAILQ_NEXT(addr, link);
	    free( addr );
	    addr = next;
	} while ( addr );	
    }
    if ( !STAILQ_EMPTY( &(lease->dns) ) )
    {
	IPv6_address *addr = STAILQ_FIRST( &(lease->dns) ), *next;
	do
	{
	    next = STAILQ_NEXT(addr, link);
	    free( addr );
	    addr = next;
	} while ( addr );
    }
    if ( lease->search_list )
	free(lease->search_list);
    if ( lease->if_name )
	free(lease->if_name);
    if ( lease->client_duid )
	free(lease->client_duid);
    if ( lease->server_duid )
	free(lease->server_duid);
    free(lease);
}

#define DHC_PAD( type, total_length ) ( ( sizeof(type) - (total_length  & (sizeof(type)-1) ) ) & (sizeof(type)-1) )

int dhcpv6_pack_lease( DHCPv6_lease* lease, uint8_t* buf, uint32_t len)
{
    if( lease == 0 ) return 0;
    
    uint8_t *bufp = buf	;

    if( buf != 0 )
	*((uint32_t*)buf) = len;

    bufp += sizeof(uint32_t);

    if( buf != 0 )
	memcpy(bufp, lease, sizeof(DHCPv6_lease));
    
    bufp += sizeof( DHCPv6_lease ) + DHC_PAD(uint32_t, sizeof(DHCPv6_lease));
    
    uint32_t *n_addrs = (uint32_t*)bufp;

    bufp += sizeof( uint32_t);

    if ( !STAILQ_EMPTY( &(lease->addresses) ) )
    {
	DHCPv6_lease_address *addr = 0;
	if( buf )
	    *n_addrs = 0;
	STAILQ_FOREACH( addr, &(lease->addresses), link )
	{
	    if( buf )
	    {
		++(*n_addrs);
		memcpy(bufp, addr, sizeof(DHCPv6_lease_address));
	    }
	    bufp +=  sizeof(DHCPv6_lease_address) + DHC_PAD(uint32_t, sizeof(DHCPv6_lease_address));
	}
    }
    n_addrs = (uint32_t*)bufp;
    bufp += sizeof(uint32_t);
    if ( !STAILQ_EMPTY( &(lease->dns) ) )
    {
	IPv6_address *addr;
	
	if( buf )
	    *n_addrs = 0;
	STAILQ_FOREACH( addr, &(lease->dns), link )
	{
	    if(buf)
	    {
		++(*n_addrs);
		memcpy(bufp, addr, sizeof(IPv6_address));
	    }
	    bufp += sizeof(IPv6_address) + DHC_PAD(uint32_t,sizeof(IPv6_address));
	}
    }

    uint32_t search_list_len = lease->search_list ? strlen(lease->search_list) : 0,
	if_name_len = lease->if_name ? strlen(lease->if_name) : 0;
	
    if(buf)
	*((uint32_t*)bufp) = search_list_len;
    bufp += sizeof(uint32_t);
    if( search_list_len )
    {
	if(buf)
	    memcpy(bufp, lease->search_list, search_list_len);
	bufp += search_list_len + DHC_PAD(uint32_t,search_list_len);
    }
    if(buf)
	*((uint32_t*)bufp) = if_name_len;
    bufp += sizeof(uint32_t);
    if( if_name_len )
    {
	if( buf )
	    memcpy(bufp, lease->if_name, if_name_len);
	bufp += if_name_len + DHC_PAD(uint32_t, if_name_len);
    }
    if( buf )
	memcpy(bufp, lease->client_duid, lease->client_duid_len);
    bufp += lease->client_duid_len;
    if( buf )
	memcpy(bufp, lease->server_duid, lease->server_duid_len);
    bufp += lease->server_duid_len;
    return (bufp - buf);
}

DHCPv6_lease *dhcpv6_unpack_lease( uint8_t* buf )
{
    if ( buf == 0 )
	return  0;

    uint8_t *bufp = buf;
    uint32_t len = *((uint32_t*) bufp);
   
    if( len < sizeof(DHCPv6_lease) )
	return 0L;

    bufp += sizeof(uint32_t);

    DHCPv6_lease *lease;
    
    lease = calloc(1, sizeof(DHCPv6_lease));
	
    memcpy(lease, bufp, sizeof( DHCPv6_lease ) );
    
    lease->search_list = 0;
    lease->if_name = 0;
    lease->client_duid=0;
    lease->server_duid=0;
    STAILQ_INIT(&(lease->addresses));
    STAILQ_INIT(&(lease->dns));

    bufp += sizeof( DHCPv6_lease ) + DHC_PAD(uint32_t, sizeof(DHCPv6_lease));
    
    uint32_t *n_addrs =(uint32_t*) bufp;
    
    bufp += sizeof(uint32_t);

    if ( *n_addrs )
    {
	DHCPv6_lease_address *addr=0;
	int n = *n_addrs;
	for(; n ; --n)
	{
	    addr = calloc(1, sizeof(DHCPv6_lease_address));
	    memcpy(addr, bufp, sizeof(DHCPv6_lease_address));
	    memset(&(addr->link),'\0',sizeof(addr->link));
	    STAILQ_INSERT_TAIL(&(lease->addresses), addr, link);
	    bufp += sizeof(DHCPv6_lease_address) + DHC_PAD(uint32_t, sizeof(DHCPv6_lease_address));	    
	}
    }
    n_addrs = (uint32_t*)bufp;
    bufp += sizeof(uint32_t);
    if( *n_addrs )
    {
	IPv6_address *addr=0;
	int n = *n_addrs;
	for(; n ; --n)
	{
	    addr = calloc(1, sizeof(IPv6_address));
	    memcpy(addr, bufp, sizeof(IPv6_address));
	    memset(&(addr->link),'\0',sizeof(addr->link));
	    STAILQ_INSERT_TAIL(&(lease->dns), addr, link);
	    bufp += sizeof(IPv6_address) + DHC_PAD(uint32_t, sizeof(IPv6_address));	    
	}	
    }
    
    uint32_t search_list_len = *((uint32_t*)bufp);
    
    bufp += sizeof(uint32_t);

    if( search_list_len )
    {
	lease->search_list = calloc(1, search_list_len);
	memcpy( lease->search_list, bufp, search_list_len );
	bufp += search_list_len + DHC_PAD(uint32_t,search_list_len);
    }
    
    uint32_t if_name_len = *((uint32_t*)bufp);

    bufp += sizeof(uint32_t);
    
    if( if_name_len )
    {
	lease->if_name = calloc(1, if_name_len);
	memcpy(lease->if_name, bufp, if_name_len);
	bufp += if_name_len + DHC_PAD(uint32_t, if_name_len);
    }
    
    if( lease->client_duid_len )
    {
	lease->client_duid = calloc(1, lease->client_duid_len);
	memcpy(lease->client_duid, bufp, lease->client_duid_len);
	bufp += lease->client_duid_len;
    }

    if( lease->server_duid_len )
    {
	lease->server_duid = calloc(1, lease->server_duid_len);
	memcpy(lease->server_duid, bufp, lease->server_duid_len);
	bufp += lease->server_duid_len;
    }
    
    return lease;
}
