/* libdhcp.c
 *
 *  libdhcp, a minimal interface to the ISC dhcp IPv4 libdhcp4client library,
 *  and to the dhcpv6 DHCPv6 client libdhcp6client library. 
 *
 * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of
 * the GNU General Public License v.2.  This program is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY expressed or
 * implied, including the implied warranties of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.  You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA. Any Red Hat trademarks that are incorporated in the source code or
 * documentation are not subject to the GNU General Public License and may
 * only be used or replicated with the express permission of Red Hat, Inc.
 *
 * Red Hat Author(s): Jason Vas Dias
 *                    David Cantrell
 */
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <sys/syslog.h>
#include <libdhcp.h>
#include <stdio.h>
#include <malloc.h>

LIBDHCP_Control 
*libdhcp_control_new
(   LIBDHCP_Callback   	   callback,
    LIBDHCP_Capability     dhc_cap, 
    time_t                 timeout,
    void                  *arg,
    LIBDHCP_Error_Handler  error_handler,
    uint8_t                log_level
)   /* creates new LIBDHCP_Control handle */
{
    LIBDHCP_Control *dhc = calloc( 1, sizeof( LIBDHCP_Control ) ); 
    if ( dhc == 0L )
	return 0L;
    dhc->capability = dhc_cap;
    dhc->callback = callback;
    dhc->timeout = timeout;
    dhc->arg = arg;
    dhc->eh = error_handler;
    dhc->log_level = log_level;
    return dhc;
}

void libdhcp_control_free(LIBDHCP_Control *dhc) {
	free(dhc);
}

extern char **environ;

int libdhcp_call_client 
(   LIBDHCP_Control *control,
    DHCP_Client client,
    ...
)
{
    va_list ap;
    int  argc=0;
    char *argv [ 32 ], **argvp=argv, *p;

    va_start(ap, client);
        
    while( (argc < 32) && ((p = va_arg(ap, char*) ) != 0 ) )
    {
	*(argvp++)=p;
	argc++;
    }
    *argvp = 0;
    va_end(ap);
    
    return (*client)(control, argc, argv, environ);
}    


char *libdhcp_state_string( DHCP_State s, char *buf )
{
    static char sbuf[32];
    char *p = buf ? &(buf[0]) : &(sbuf[0]);
    
    sprintf
    (   p, "%s",
	(s == DHC4_NBI)
	?"DHC4_NBI"
	:(s == DHC4_PREINIT)
	?"DHC4_PREINIT"
	:(s == DHC4_BOUND)
	?"DHC4_BOUND"
	:(s == DHC4_RENEW)
	?"DHC4_RENEW"
	:(s == DHC4_REBOOT)
	?"DHC4_REBOOT"
	:(s == DHC4_REBIND)
	?"DHC4_REBIND"
	:(s == DHC4_STOP)
	?"DHC4_STOP"
	:(s == DHC4_MEDIUM)
	?"DHC4_MEDIUM"
	:(s == DHC4_TIMEOUT)
	?"DHC4_TIMEOUT"
	:(s == DHC4_FAIL)
	?"DHC4_FAIL"
	:(s == DHC4_EXPIRE)
	?"DHC4_EXPIRE"
	:(s == DHC4_RELEASE)
	?"DHC4_RELEASE"
	:(s == DHC_TIMEDOUT)
	?"DHC_TIMEDOUT"
	:(s == DHC6_BOUND)
	?"DHC6_BOUND"
	:(s == DHC6_REBIND)
	?"DHC6_REBIND"
	:(s == DHC6_RELEASE)
	?"DHC6_RELEASE"
        :"DHC_INVALID"
    );

    return p;
}

int 
libdhcp_stderr_logger
(   struct libdhcp_control_s *ctl,
    int priority,  /* ala syslog(3): LOG_EMERG=0 - LOG_DEBUG=7 (+ LOG_FATAL=8 : finished -> 1)   */
    const char *fmt,
    va_list ap
)
{
    struct tm tm;
    int len=0;
    if ( (priority != LOG_FATAL) && ctl && (priority > ctl->log_level) )
	return 0;
    time_t t=time(0);
    tm=*localtime(&t);	
    fprintf(stderr, "%.4u-%.2u-%.2u_%.2u:%.2u:%.2u libdhcp [%d] %s:",
	   tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec,
	   getpid(),
	   (priority == LOG_FATAL)
	   ? "FATAL"
	   : (priority == LOG_ERR)
	   ? "ERROR"
	   : (priority == LOG_INFO)
	   ? "INFO"
	   : (priority == LOG_DEBUG)
	   ? "DBG"
	   : ""
	);
    if(priority == LOG_FATAL)
	if( ctl )
	    ctl->finished = 1;
    len+=vprintf(fmt,ap);
    printf("\n");
    fflush(stderr);
    return len;
}

int
libdhcp_syslogger
(   struct libdhcp_control_s *ctl,
    int priority,  /* ala syslog(3): LOG_EMERG=0 - LOG_DEBUG=7 (+ LOG_FATAL=8 : finished -> 1)   */
    const char *fmt,
    va_list ap
)
{
    if ( (priority != LOG_FATAL) && ctl && (priority > ctl->log_level) )
	return 0;
    if(priority == LOG_FATAL)
    {
	if(ctl)
	    ctl->finished = 1;	
	priority = LOG_ERR;
    }
    vsyslog( priority, fmt, ap );
    return 1;
}
