/** @file nic.c
 *
 *  Network Interface Configurator (NIC) for libdhcp.
 *
 *  Provides facilities for configuring network interfaces, addresses,
 *  and routes.
 * 
 */
/*
 * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of
 * the GNU General Public License v.2.  This program is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY expressed or
 * implied, including the implied warranties of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.  You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA. Any Red Hat trademarks that are incorporated in the source code or
 * documentation are not subject to the GNU General Public License and may
 * only be used or replicated with the express permission of Red Hat, Inc.
 *
 * Red Hat Author(s): Jason Vas Dias
 *                    David Cantrell
 */

#define _GNU_SOURCE 1

#include <sys/types.h>
#include <unistd.h>
#include <nic.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/syslog.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <search.h>
#include <time.h>
#include <errno.h>

#include <libdhcp.h>

/* libnl Netlink API headers: */

#include <netlink/netlink.h>
#include <netlink/netlink-kernel.h>
#include <netlink/rtnetlink-kernel.h>
#include <netlink/msg.h>
#include <netlink/attr.h>
#include <netlink/utils.h>
#include <netlink/addr.h>
#include <netlink/route/rtnl.h>
#include <netlink/route/link.h>
#include <netlink/route/addr.h>
#include <netlink/route/route.h>
extern int nl_get_errno(void);

#define eprintf(level, fmt, arg...) ({if (nh && nh->eh) (nh->eh)(nh, (level), (fmt), ##arg ); })

/**
 * NIC library handle
 */
struct nlh_s
{
    struct nl_handle *nl;   /* netlink handle */

    void *nic_name_tree;    /* glibc search(3) btrees */
    void *nic_index_tree;
    void *nic_foreach_arg;  /* nic_foreach() callback arg */

    void *nic_addr_tree;
    void *nic_addr_foreach_arg;      

    void *nic_route_tree;
    void *nic_route_foreach_arg;

    int               ll;   /* log level      */    
    NIC_Error_Handler_t    eh;       /* error handler / logger */
    NIC_VA_Error_Handler_t va_eh;    /* underlying va_list logger (libdhcp) */
    void                  *va_eh_arg;/* va_list logger 1st argument */
};

NLH_t nic_open(NIC_Error_Handler_t eh)
{
    NLH_t nh;
    
    if (!(nh = calloc(1, sizeof(struct nlh_s))))
	goto nic_open_fail;
    
    if (eh) {
        nh->ll = NIC_ERR;
        nh->eh = eh;
    }
    if (!(nh->nl = nl_handle_alloc())) {
	eprintf(NIC_FATAL, "cannot allocate NIC library handle: %m" );
	goto nic_open_fail;
    }
    
    if (nl_connect(nh->nl, NETLINK_ROUTE) < 0 ) {
        eprintf(NIC_FATAL, "cannot connect netlink socket: %m" );
        goto nic_open_fail;
    }
    
    return nh;

nic_open_fail:
    if (nh) {
        if (nh->nl) {
            nl_handle_destroy(nh->nl);
            nh->nl = NULL;
        }
        free(nh);
    }
    return NULL;
}

static void tdestroy_nil(void *n) {}

void nic_close(struct nlh_s **nhp)
{
    struct nlh_s *nh;

    if (!nhp || !*nhp)
        return;
    nh = *nhp;

    if (nh->nl) {
	nl_close(nh->nl);       
	nl_handle_destroy(nh->nl);
        nh->nl = NULL;
    }

    if (nh->nic_name_tree) {
	tdestroy(nh->nic_name_tree, free);
	if (nh->nic_index_tree)
	    tdestroy(nh->nic_index_tree, tdestroy_nil);
	nh->nic_name_tree = 0;
	nh->nic_index_tree = 0;
    } else if (nh->nic_index_tree) {
	tdestroy(nh->nic_index_tree, free);
	if (nh->nic_name_tree)
	    tdestroy(nh->nic_name_tree, tdestroy_nil);
	nh->nic_name_tree = 0;
	nh->nic_index_tree = 0;
    }
    
    if (nh->nic_addr_tree) {
	tdestroy(nh->nic_addr_tree, nic_addr_free);
	nh->nic_addr_tree = 0;
    }

    if (nh->nic_route_tree)
    {
	tdestroy(nh->nic_route_tree, nic_route_free);
	nh->nic_route_tree = 0;
    }
    
    free(nh);
    *nhp = NULL;
}


/* default error handlers: */

void nic_sys_logger( NLH_t nh, NIC_Error_Level_t el, char *fmt, ... )
{
    if ( (nh == 0L) || ( el > nh->ll ) )
	return;
    va_list va;
    
    va_start(va, fmt);
    
    if( nh->va_eh )
	nh->va_eh ( nh->va_eh_arg, el, fmt, va );
    else 
	vsyslog( el, fmt, va );

    va_end(va);

    if ( el == LOG_FATAL )
	nic_close(&nh);
}

void nic_stderr_logger( NLH_t nh, NIC_Error_Level_t el, char *fmt, ... )
{
    if ( (nh == 0L) || ( el > nh->ll ) )
	return;

    va_list va;
    
    va_start(va, fmt);
    if( nh->va_eh )
	nh->va_eh ( nh->va_eh_arg, el, fmt, va );
    else 
    {
	vfprintf(stderr, fmt, va);
	fprintf(stderr,"\n");
	fflush(stderr);
    }
    va_end(va);
}

NIC_Res_t nic_set_loglevel( NLH_t nh, NIC_Error_Level_t ll )
{
    if ( nh == 0 )
	return NIC_FAIL;

    nh -> ll = ll;
    
    return NIC_OK;
}

extern
void nic_set_logger(NLH_t nh, NIC_Error_Handler_t eh)
{
    nh -> eh = eh;
}

extern
void nic_set_va_logger(NLH_t nh, NIC_VA_Error_Handler_t va_eh, void *va_eh_arg)
{
    nh -> va_eh = va_eh;
    nh -> eh = nic_stderr_logger;
}

#ifndef IFQDISCSIZ
#define IFQDISCSIZ 32
#endif

struct nic_s
{
    NLH_t             nh;      /**< nic handle */

    struct nic_link
    {
	struct nlmsghdr   hdr;
	struct ifinfomsg  ifi;
        /* 
	 *struct ifinfomsg
	 *{
	 *
	 *	unsigned char	ifi_family;
	 *	unsigned char	__ifi_pad;
	 *	unsigned short	ifi_type;
	 *	int		ifi_index;
	 *	unsigned	ifi_flags;
	 *	unsigned	ifi_change;
	 *};
	 */
	
	ip_addr_t       addr;

	ip_addr_t       broadcast;

	char		name[IFNAMSIZ];	
	char		qdisc[IFQDISCSIZ];	
	uint32_t 	mtu;
	uint32_t	link;
	uint32_t	txqlen;
	uint32_t	weight;
	uint32_t	master;  
	uint32_t        cost;
	uint32_t        priority;
	uint32_t        protinfo;
	struct rtnl_link_stats stats;
	struct rtnl_link_ifmap ifmap;

	enum nic_mask_e
	{
	    NIC_LINK_ADDRESS   =   1,
	    NIC_LINK_BROADCAST =   2,
	    NIC_LINK_NAME      =   4,
	    NIC_LINK_MTU       =   8,
	    NIC_LINK_LINK      =  16,
	    NIC_LINK_QDISC     =  32,
	    NIC_LINK_STATS     =  64,
	    NIC_LINK_COST      = 128,
	    NIC_LINK_PRIORITY  = 256,
	    NIC_LINK_MASTER    = 512,
	    NIC_LINK_WIRELESS  =1024, /**<unhandled for now */
	    NIC_LINK_PROTINFO  =2048,
	    NIC_LINK_TXQLEN    =4096,
	    NIC_LINK_IFMAP     =8192,	    
	    NIC_LINK_WEIGHT    =16384,
	    NIC_LINK_OPERSTATE =32768,
	    NIC_LINK_LINKMODE  =65536
	} have, change;
    } l;
};

static int nic_name_comparator( const void *p1, const void *p2 )
{
    const struct nic_s *n1 = p1, *n2 = p2;
    return strcmp(n1->l.name, n2->l.name);
}

static int nic_index_comparator( const void *p1, const void *p2 )
{
    const struct nic_s *n1 = p1, *n2 = p2;
    return ( (n1->l.ifi.ifi_index == n2->l.ifi.ifi_index)
	     ? 0
	     : (n1->l.ifi.ifi_index > n2->l.ifi.ifi_index)
	       ?  1
	       : -1
	   );
}

static int nic_get_links(NLH_t nh, char *if_name, int if_index) {
    char *buf = NULL;
    char ipbuf[INET6_ADDRSTRLEN+1];
    struct nl_cache *cache = NULL;
    struct rtnl_link *rlink = NULL;
    struct nl_addr *addr = NULL;
    int nitems = 0, i = 1;
    int find_name = 0, find_index = 0, nlinks = -1;
    struct nic_s *nic = NULL, **nnic = NULL, **inic = NULL;

    if ((cache = rtnl_link_alloc_cache(nh->nl)) == NULL) {
        eprintf(NIC_FATAL, "%s (%d): %s\n",
                __func__, __LINE__, strerror(errno));
        return nlinks;
    }

    if (if_name && *if_name) {
        eprintf(NIC_DEBUG, "looking for name %s\n", if_name);
        find_name = -1;
    } else if (if_index != -1) {
        eprintf(NIC_DEBUG, "looking for index %d\n", if_index);
        find_index = -1;
    }

    nitems = nl_cache_nitems(cache);

    while (i <= nitems) {
        if ((rlink = rtnl_link_get(cache, i)) == NULL) {
            eprintf(NIC_FATAL, "%s (%d): %s\n",
                    __func__, __LINE__, strerror(errno));
            break;
        }

        if (find_name) {
            if (strcmp(if_name, rtnl_link_get_name(rlink))) {
                rtnl_link_put(rlink);
                i++;
                continue;
            }
        } else if (find_index) {
            if (if_index != rtnl_link_get_ifindex(rlink)) {
                rtnl_link_put(rlink);
                i++;
                continue;
            }
        }

        if ((nic = calloc(1, sizeof(*nic))) == NULL) {
            rtnl_link_put(rlink);
            eprintf(NIC_FATAL, "%s (%d): %s\n",
                    __func__, __LINE__, strerror(errno));
            break;
        }

        nic->nh = nh;

        memset(&(nic->l.ifi), 0, sizeof(nic->l.ifi));
        nic->l.ifi.ifi_family = rtnl_link_get_family(rlink);
        nic->l.ifi.ifi_index = rtnl_link_get_ifindex(rlink);
        nic->l.ifi.ifi_flags = rtnl_link_get_flags(rlink);

        memset(nic->l.name, '\0', sizeof(nic->l.name));
        buf = rtnl_link_get_name(rlink);
        if (buf == NULL) {
            eprintf(NIC_DEBUG, "didn't get name\n");
        } else {
            eprintf(NIC_DEBUG, "got name: %s", buf);
            strncpy(nic->l.name, buf, sizeof(nic->l.name));
            nic->l.have |= NIC_LINK_NAME;
        }

        if ((addr = rtnl_link_get_addr(rlink)) == NULL) {
            rtnl_link_put(rlink);
            eprintf(NIC_FATAL, "%s (%d): %s\n",
                    __func__, __LINE__, strerror(errno));
            break;
        }

        memset(&ipbuf, '\0', sizeof(ipbuf));

        if (nl_addr2str(addr, ipbuf, INET6_ADDRSTRLEN) == NULL) {
            rtnl_link_put(rlink);
            eprintf(NIC_FATAL, "%s (%d): %s\n",
                    __func__, __LINE__, strerror(errno));
            break;
        }

        if (ipbuf && *ipbuf) {
            nic->l.addr = ip_addr_text(ipbuf);
            nic->l.have |= NIC_LINK_ADDRESS;
        }

        if ((addr = rtnl_link_get_broadcast(rlink)) == NULL) {
            rtnl_link_put(rlink);
            eprintf(NIC_FATAL, "%s (%d): %s\n",
                    __func__, __LINE__, strerror(errno));
            break;
        }

        memset(&ipbuf, '\0', sizeof(ipbuf));

        if (nl_addr2str(addr, ipbuf, INET6_ADDRSTRLEN) == NULL) {
            rtnl_link_put(rlink);
            eprintf(NIC_FATAL, "%s (%d): %s\n",
                    __func__, __LINE__, strerror(errno));
            break;
        }

        if (ipbuf && *ipbuf) {
            nic->l.broadcast = ip_addr_text(ipbuf);
            nic->l.have |= NIC_LINK_BROADCAST;
        }

        nic->l.txqlen = rtnl_link_get_txqlen(rlink);
        if (nic->l.txqlen) {
            nic->l.have |= NIC_LINK_TXQLEN;
        }

        nic->l.mtu = rtnl_link_get_mtu(rlink);
        if (nic->l.mtu) {
            nic->l.have |= NIC_LINK_MTU;
        }

        nic->l.link = rtnl_link_get_link(rlink);
        if (nic->l.link) {
            nic->l.have |= NIC_LINK_LINK;
        }

        nic->l.weight = rtnl_link_get_weight(rlink);
        if (nic->l.weight) {
            nic->l.have |= NIC_LINK_WEIGHT;
        }

        nic->l.master = rtnl_link_get_master(rlink);
        if (nic->l.master) {
            nic->l.have |= NIC_LINK_MASTER;
        }

        buf = rtnl_link_get_qdisc(rlink);
        if (buf != NULL) {
            strncpy(nic->l.qdisc, buf, sizeof(nic->l.qdisc));
            nic->l.have |= NIC_LINK_QDISC;
        }

        rtnl_link_put(rlink);
        i++;

        nnic = tfind(nic, &nh->nic_name_tree, nic_name_comparator);
        if (!nnic) {
            if (!(nnic = tsearch(nic, &(nh->nic_name_tree),
                                 nic_name_comparator))) {
                free(nic);
                break;
            }

            if (!(inic = tsearch(nic, &(nh->nic_index_tree),
                                 nic_index_comparator))) {
                free(nic);
                break;
            }
        }

        if (nic != *nnic) {
            free(nic);
            nic = *nnic;
        }

        if ((find_index && nic->l.ifi.ifi_index == if_index) ||
            (find_name && !strcmp(nic->l.name, if_name))) {
            break;
        }
    }

    nl_cache_destroy_and_free(cache);
    return nitems;
}

NIC_t nic_by_name(NLH_t nh, char *if_name)
{    
    struct nic_s snic, *nic=0, **pnic=0;

    if (!nh->nic_name_tree)
        nic_get_links(nh, "", -1);

    if( (if_name == 0L) || ( *if_name == '\0' ) )
    {
        eprintf(NIC_FATAL, "nic_by_name: invalid nil argument" );
	return( 0L );	
    }
   
    strncpy(&(snic.l.name[0]),if_name, IFNAMSIZ);

    if ( (pnic = tfind(&snic, &(nh->nic_name_tree), nic_name_comparator)) != 0L )
	return *pnic;

    if( ( nic_get_links(nh, if_name, -1) <= 0 ) 
      ||((pnic = tfind(&snic, &(nh->nic_name_tree), nic_name_comparator)) == 0L)
      )
    {
        eprintf(NIC_ERR, "nic_by_name: no interface named %s found", if_name );
	return(0L);
    }else
	nic = *pnic;
    
    return nic;
}

NIC_t nic_by_index(NLH_t nh, int16_t if_index)
{    
    struct nic_s snic, *nic=0, **pnic=0;

    /* nic_index_tree will be populated at the same time */
    if (!nh->nic_name_tree)
        nic_get_links(nh, "", -1);

    if( if_index == -1 )
    {
        eprintf(NIC_FATAL, "nic_by_index: invalid -1 argument" );
	return( 0L );	
    }
   
    snic.l.ifi.ifi_index = if_index;

    if ( (pnic = tfind(&snic, &(nh->nic_index_tree), nic_index_comparator)) != 0L )
	return *pnic;


    if( ( nic_get_links(nh, "", if_index ) <= 0 ) 
      ||((pnic = tfind(&snic, &(nh->nic_index_tree), nic_index_comparator)) == 0L)
      )
    {
        eprintf(NIC_ERR, "nic_by_index: no interface with index %d found", if_index );
	return(0L);
    }else
	nic = *pnic;
    
    return nic;
}

struct nic_cbarg
{
    NIC_handler_t cb;
    void        *arg;
};

static void nic_twalker( const void *p, const VISIT which, const int depth )
{
    struct nic_s *nic, *const*npp=p;
    struct nic_cbarg *cb;

    if( (npp == 0L)    || ((nic = *npp) == 0L)
      ||(nic->nh == 0) || (nic->nh->nic_foreach_arg == 0 )
      ||( (which != postorder) && (which != leaf) )
      ) return;

    cb = nic->nh->nic_foreach_arg;
 
    cb->cb( nic->nh, nic, cb->arg );       
}

void nic_foreach(NLH_t nh, NIC_handler_t handler,  void *arg)
{
    struct nic_cbarg cb = { handler, arg };
    struct nic_s snic;

    if ( nh->nic_foreach_arg )
	return;

    nh->nic_foreach_arg = &cb;
    memset(&snic, '\0', sizeof(struct nic_s));
    snic.l.ifi.ifi_index = -1;
    snic.nh = nh;

    /* Ensure we have ALL current links in the tree */
    nic_get_links(nh, "", -1 ) ;

    if( nh->nic_name_tree )
	twalk( nh->nic_name_tree, nic_twalker);

    nh->nic_foreach_arg = 0L;
}

char* nic_get_name( NIC_t nic )
{
    if( nic == 0 ) return 0;
	
    return &(nic->l.name[0]);
}

int16_t nic_get_index( NIC_t nic )
{
    if ( nic == 0 ) return -1;

    return nic->l.ifi.ifi_index ;
}

ip_addr_t nic_get_link_addr(NIC_t nic)
{
    ip_addr_t ip;
    memset(&ip, '\0', sizeof(ip));
    if( nic == 0L )
	return ip;
    return nic->l.addr;
}

ip_addr_t nic_get_link_broadcast(NIC_t nic)
{
    ip_addr_t ip;
    memset(&ip, '\0', sizeof(ip));
    if( nic == 0L )
	return ip;
    return nic->l.broadcast;
}

uint32_t nic_get_flags( NIC_t nic )
{
    if(nic == 0L)
	return 0;
    return nic->l.ifi.ifi_flags;
}

NIC_Res_t nic_update(NIC_t nic) {
    NLH_t nh;
    struct nl_cache *cache = NULL;
    struct rtnl_link *link = NULL;
    struct rtnl_link *request = NULL;
    struct nl_msg *msg = NULL;
    int ret = 0;

    nh = nic->nh;

    if (nic == NULL) {
        return NIC_FAIL;
    }

    if ((cache = rtnl_link_alloc_cache(nh->nl)) == NULL) {
        return NIC_FAIL;
    }

    if ((link = rtnl_link_get(cache, nic->l.ifi.ifi_index)) == NULL) {
        nl_cache_destroy_and_free(cache);
        return NIC_FAIL;
    }

    request = rtnl_link_alloc();
    rtnl_link_set_flags(request, nic_get_flags(nic) | NLM_F_REPLACE |
                                 NLM_F_REQUEST | NLM_F_ACK |
                                 NLM_F_MATCH | NLM_F_ATOMIC);

    rtnl_link_set_ifindex(request, nic->l.ifi.ifi_index);
    rtnl_link_set_family(request, nic->l.ifi.ifi_family);
    rtnl_link_set_master(request, nic->l.master);

    if (nic->l.have & NIC_LINK_NAME) {
        rtnl_link_set_name(request, nic->l.name);
    }

    if (nic->l.have & NIC_LINK_QDISC) {
        rtnl_link_set_qdisc(request, nic->l.qdisc);
    }

    if (nic->l.have & NIC_LINK_LINK) {
        rtnl_link_set_link(request, nic->l.link);
    }

    if (nic->l.have & NIC_LINK_MTU) {
        rtnl_link_set_mtu(request, nic->l.mtu);
    }

    if (nic->l.have & NIC_LINK_TXQLEN) {
        rtnl_link_set_txqlen(request, nic->l.txqlen);
    }

    if (nic->l.have & NIC_LINK_WEIGHT) {
        rtnl_link_set_weight(request, nic->l.weight);
    }

    msg = rtnl_link_build_change_request(link, request, 0);
    ret = nl_send_auto_complete(nh->nl, nlmsg_hdr(msg));
    nlmsg_free(msg);
    rtnl_link_put(link);
    rtnl_link_put(request);
    nl_cache_destroy_and_free(cache);

    if (ret < 0) {
        eprintf(NIC_ERR, "nic_update failure (line %d): %s (error code: %d)",
                __LINE__, nl_geterror(), ret);
        return NIC_FAIL;
    } else {
        return NIC_OK;
    }
}

void nic_set_flags( NIC_t nic, uint32_t flags )
{
    if(nic == 0L) return;
	
    nic->l.ifi.ifi_flags = flags;
}

uint32_t nic_get_mtu( NIC_t nic )
{
    if((nic == 0L) || ((nic->l.have & NIC_LINK_MTU)==0))
	return 0;
    return nic->l.mtu;
}

void nic_set_mtu( NIC_t nic, uint32_t mtu )
{
    if( nic == 0 ) return ;
    nic->l.mtu = mtu;
    nic->l.change |= NIC_LINK_MTU;
}

char *nic_get_qdisc( NIC_t nic )
{
    if((nic == 0) || ((nic->l.have & NIC_LINK_QDISC)==0))
	return 0;
    return &(nic->l.qdisc[0]);    
}

extern
void nic_set_qdisc( NIC_t nic, char *qdisc )
{
    if(nic == 0) return ;
    strncpy(&(nic->l.qdisc[0]), qdisc, IFQDISCSIZ);
    nic->l.change |= NIC_LINK_QDISC;
}

uint32_t nic_get_txqlen( NIC_t nic )
{
    if((nic == 0L) || ((nic->l.have & NIC_LINK_TXQLEN)==0))
	return 0;
    return nic->l.txqlen;
}

void nic_set_txqlen( NIC_t nic, uint32_t txq )
{
    if(nic == 0L) return;
    nic->l.txqlen = txq;
    nic->l.change |= NIC_LINK_TXQLEN;
}

uint32_t nic_get_link( NIC_t nic )
{
    if((nic == 0L) || ((nic->l.have & NIC_LINK_LINK)==0))
	return 0;
    return nic->l.link;
}

void nic_set_link( NIC_t nic, uint32_t link )
{
    if( nic == 0L) return;
    nic->l.link = link;
    nic->l.change |= NIC_LINK_LINK;
}

uint32_t nic_get_weight( NIC_t nic )
{
    if((nic == 0L) || ((nic->l.have & NIC_LINK_WEIGHT)==0))
	return 0;
    return nic->l.weight;
}

void nic_set_weight( NIC_t nic, uint32_t weight )
{
    if( nic == 0L) return;
    nic->l.weight = weight;
    nic->l.change |= NIC_LINK_WEIGHT;
}

uint32_t nic_get_master( NIC_t nic )
{
    if((nic == 0L) || ((nic->l.have & NIC_LINK_MASTER)==0))
	return 0;
    return nic->l.master;
}

void nic_set_master( NIC_t nic, uint32_t master )
{
    if( nic == 0L) return;
    nic->l.master = master;
    nic->l.change |= NIC_LINK_MASTER;
}

uint32_t nic_get_cost( NIC_t nic )
{
    if((nic == 0L) || ((nic->l.have & NIC_LINK_COST)==0))
	return 0;
    return nic->l.cost;
}

void nic_set_cost( NIC_t nic, uint32_t cost )
{
    if( nic == 0L) return;
    nic->l.cost = cost;
    nic->l.change |= NIC_LINK_COST;
}

uint32_t nic_get_priority( NIC_t nic )
{
    if((nic == 0L) || ((nic->l.have & NIC_LINK_PRIORITY)==0))
	return 0;
    return nic->l.priority;
}

void nic_set_priority( NIC_t nic, uint32_t priority )
{
    if( nic == 0L) return;
    nic->l.priority = priority;
    nic->l.change |= NIC_LINK_PRIORITY;
}

uint32_t nic_get_protinfo( NIC_t nic )
{
    if((nic == 0L) || ((nic->l.have & NIC_LINK_PROTINFO)==0))
	return 0;
    return nic->l.protinfo;
}

void nic_set_protinfo( NIC_t nic, uint32_t protinfo )
{
    if( nic == 0L) return;
    nic->l.protinfo = protinfo;
    nic->l.change |= NIC_LINK_PROTINFO;
}

struct rtnl_link_stats nic_get_stats(NIC_t nic)
{
    struct rtnl_link_stats st;
    memset(&st, '\0', sizeof(struct rtnl_link_stats));
    if((nic == 0L) || ((nic->l.have & NIC_LINK_STATS)==0))
	return st;    
    return nic->l.stats;
}

struct rtnl_link_ifmap nic_get_ifmap(NIC_t nic)
{
    struct rtnl_link_ifmap ifm;
    memset(&ifm, '\0', sizeof(struct rtnl_link_ifmap));
    if((nic == 0L) || ((nic->l.have & NIC_LINK_IFMAP)==0))
	return ifm;    
    return nic->l.ifmap;
}

struct nic_ip_address_s 
{
    NLH_t nh;
    

    struct nic_addr_s
    {
	struct nlmsghdr   hdr;
	struct ifaddrmsg  ifa;
        /*
	 *struct ifaddrmsg
	 *{
	 *   unsigned char	ifa_family;
	 *   unsigned char	ifa_prefixlen;
	 *   unsigned char	ifa_flags;
	 *   unsigned char	ifa_scope;
	 *   int	        ifa_index;
	 *};
         */
	uint32_t   index;
	uint32_t   flags;
	ip_addr_t  addr;
	ip_addr_t  broadcast;
	ip_addr_t  anycast;
	ip_addr_t  multicast;
	ip_addr_t  local;
	struct ifa_cacheinfo cacheinfo;
	char       label [ IFNAMSIZ ];

	enum ifa_e
	{	    
	    NIC_ADDR_ADDRESS   =   1,
	    NIC_ADDR_LOCAL     =   2,
	    NIC_ADDR_LABEL     =   4,
	    NIC_ADDR_BROADCAST =   8,
	    NIC_ADDR_ANYCAST   =  16,
	    NIC_ADDR_CACHEINFO =  32,
	    NIC_ADDR_MULTICAST =  64
	} have, change;
	
    } nla;    
};

static struct nla_policy nic_addr_policy[IFA_MAX+1] = {
	[IFA_LABEL]	= { .type = NLA_STRING,
			    .maxlen = IFNAMSIZ },
	[IFA_CACHEINFO]	= { .minlen = sizeof(struct ifa_cacheinfo) },
};

static int nic_addr_comparator( const void *p1, const void *p2 )
{
    const struct nic_ip_address_s  *ipa1=p1, *ipa2=p2;

    if( ipa1->nla.ifa.ifa_index > ipa2->nla.ifa.ifa_index )
	return 1;
    else
    if( ipa1->nla.ifa.ifa_index < ipa2->nla.ifa.ifa_index )
	return -1;

    uint8_t 
	pfx1 = ipa1->nla.ifa.ifa_prefixlen,
	pfx2 = ipa2->nla.ifa.ifa_prefixlen;

    const ip_addr_t *ip1=0, *ip2=0;
    
    if( ipa1->nla.have & NIC_ADDR_ADDRESS )
	ip1 = &(ipa1->nla.addr);
    else
    if( ipa1->nla.have & NIC_ADDR_LOCAL )
    {
	ip1 = &(ipa1->nla.local);
	pfx1 = IP_ADDR_SIZE(ip1) * 8;
    }

    if( ipa2->nla.have & NIC_ADDR_ADDRESS )
	ip2 = &(ipa2->nla.addr);
    else
    if( ipa2->nla.have & NIC_ADDR_LOCAL )
    {
	ip2 = &(ipa2->nla.local);
	pfx2= IP_ADDR_SIZE(ip2) * 8;
    }

    if( ip1 == 0 )
	return -1;
    
    if( ip2 == 0 )
	return 1;

    uint8_t
	alen1 = IP_ADDR_SIZE(ip1),
	alen2 = IP_ADDR_SIZE(ip2);
    
    if( alen1 > alen2 )
	return 1;
    else
    if( alen1 < alen2 )
	return -1;

    if( pfx1 > pfx2 )
	return 1;
    else
    if( pfx1 < pfx2 )
	return -1;

    const uint8_t
	*ap1 = IP_ADDR(ip1),
	*ap2 = IP_ADDR(ip2);

    for(; alen1 && pfx1; alen1--, pfx1-=8, ap1++, ap2++ )
    {
	const uint8_t 
	    v1=*ap1,
	    v2=*ap2;

	if( pfx1 < 8 )
	{
	    const uint8_t 
		vb1 = v1 & ((1 << pfx1)-1),
		vb2 = v2 & ((1 << pfx1)-1);
	    return
		( vb1 > vb2 )
		? 1
		: ( vb1 < vb2 )
		  ? -1
		  : 0;
	}else
	{
	    if( v1 > v2 )
		return 1;
	    if( v1 < v2 )
		return -1;
	}
    }
    return 0;
}

static int nic_get_addresses(NLH_t nh, IPaddr_t ipa)
{
    char abuf[32];
    struct nic_ip_address_s ipaf, *addr;

    if(nh == 0)
	return 0;
    
    int n_addrs = 0;
    struct nic_addr_s *nla;    

    if( ipa )
    {
	ipaf = *ipa;
	addr = ipa;
	nla = &(ipa->nla);    
	memset(&(nla->hdr),'\0',sizeof(struct nlmsghdr));
	nla->hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_ATOMIC | NLM_F_MATCH;
    }else
    {
	memset(&ipaf, '\0', sizeof(struct nic_ip_address_s));
	ipaf.nh = nh;
	addr = &ipaf;
	nla = &(ipaf.nla);    
	addr=&ipaf;	
	nla->hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
    }
    
    nla->hdr.nlmsg_type = RTM_GETADDR;
    
    struct nl_msg *msg = nlmsg_build(&(nla->hdr));

    if( msg == 0L )
    {
        eprintf(NIC_ERR, "nic_get_addresses: out of memory" );
	return n_addrs;
    }
    
    if( nlmsg_append(msg, &(nla->ifa), sizeof(struct ifaddrmsg), 1) < 0 )
    {	
        eprintf(NIC_ERR,"nic_get_addresses: nlmsg_append failed: %d %s\n", 
		nl_get_errno(), nl_geterror());
	goto nla_put_failure;
    }

    if( addr->nla.addr.sa_family )
	NLA_PUT(msg, IFLA_ADDRESS, 
		IP_ADDR_SIZE(&(addr->nla.addr)), 
		IP_ADDR(&(addr->nla.addr))
	       );

    if ( nl_send_auto_complete( nh->nl, nlmsg_hdr(msg) ) < 0 )
    {
        eprintf(NIC_ERR,"nic_get_addresses: send failed: %d %s\n", 
		nl_get_errno(), nl_geterror());
	goto nla_put_failure;
    }

    struct nic_addr_s *buf = 0;

    if( buf == 0L )
    {
        eprintf(NIC_ERR,"nic_get_addresses: out of memory.");
	goto nla_put_failure;
    }

    struct nlmsghdr *hdr = (struct nlmsghdr *)buf;
    struct sockaddr_nl kernel;
    int rlen = 0;
    int remaining = 0;
    
    do
    {
	if ( (rlen = nl_recv( nh->nl, &kernel, (void*)&buf)) > 0 )
	{
	    if( rlen < (sizeof(struct nlmsghdr) + sizeof(struct ifaddrmsg)) )
		goto nla_return;

	    hdr = (struct nlmsghdr*)buf;
	    remaining = rlen;

	    if( hdr->nlmsg_type != RTM_NEWADDR )
		goto nla_return;
	    
	    do
	    {
		memset(nla, '\0', sizeof(struct nic_addr_s));
		
		memcpy(nla, hdr, sizeof(struct nlmsghdr));
		
		struct nl_attr *tb[__IFA_MAX];
		
		int i;
		for (i=0; i < __IFA_MAX; i++) tb[i]=0;
		int err = 
		    nlmsg_parse((struct nlmsghdr*)hdr, sizeof(struct ifaddrmsg),
				(void*)tb, IFA_MAX, nic_addr_policy);
		if( err < 0 )
		{
                    eprintf(NIC_ERR,"nic_get_addresses: nl_parse failed: %d %s",
			    err, nl_geterror());
		    goto nla_return;
		}
		
		nla->ifa = *((struct ifaddrmsg*)nlmsg_data((struct nlmsghdr*)hdr));
		
		if( tb[ IFA_ADDRESS ] )
		{
		    nla->addr = 
			ip_addr_binary 
			( nla_data((void*)tb[IFA_ADDRESS]),
			  nla_len((void*)tb[IFA_ADDRESS])
			);	

                    eprintf(NIC_DEBUG,"nic %d address: %s/%hd",
			    nla->ifa.ifa_index,
		            ip_text_addr(&nla->addr, abuf, 32),
                            nla->ifa.ifa_prefixlen);
		    nla->have |= NIC_ADDR_ADDRESS;
		}

		if( tb[ IFA_LOCAL ] )
		{
		    nla->local = 
			ip_addr_binary 
			( nla_data((void*)tb[IFA_ADDRESS]),
			  nla_len((void*)tb[IFA_ADDRESS])
			);	
		    eprintf(NIC_DEBUG,"nic %d local (PtP) address: %s",
			    nla->ifa.ifa_index,
                            ip_text_addr(&nla->local, abuf, 32));
		    nla->have |= NIC_ADDR_LOCAL;
		}

		if( tb[ IFA_BROADCAST ] )
		{
		    nla->broadcast = 
			ip_addr_binary 
			( nla_data((void*)tb[IFA_BROADCAST]),
			  nla_len((void*)tb[IFA_BROADCAST])
			);	
		    eprintf(NIC_DEBUG,"nic %d broadcast address: %s",
			    nla->ifa.ifa_index,
                            ip_text_addr(&nla->broadcast, abuf, 32));
		    nla->have |= NIC_ADDR_BROADCAST;
		}

		if( tb[ IFA_ANYCAST ] )
		{
		    nla->anycast = 
			ip_addr_binary 
			( nla_data((void*)tb[IFA_ANYCAST]),
			  nla_len((void*)tb[IFA_ANYCAST])
			);	
                    eprintf(NIC_DEBUG,"nic %d anycast address: %s",
			    nla->ifa.ifa_index,
                            ip_text_addr(&nla->anycast, abuf, 32));
		    nla->have |= NIC_ADDR_ANYCAST;
		}
		
		if( tb[ IFA_MULTICAST ] )
		{
		    nla->multicast = 
			ip_addr_binary 
			( nla_data((void*)tb[IFA_MULTICAST]),
			  nla_len((void*)tb[IFA_MULTICAST])
			);	
                    eprintf(NIC_DEBUG,"nic %d multicast address: %s",
			    nla->ifa.ifa_index,
		            ip_text_addr(&nla->multicast, abuf, 32));
		    nla->have |= NIC_ADDR_MULTICAST;
		}

		if( tb[IFA_CACHEINFO] )
		{
		    memcpy
			( &(nla->cacheinfo), 
			  nla_data((void*)tb[IFA_CACHEINFO]),
			  sizeof(struct ifa_cacheinfo)
			);
		    eprintf(NIC_DEBUG,"nic %d have cacheinfo", 
                            nla->ifa.ifa_index);
		    nla->have |= NIC_ADDR_CACHEINFO;
		}

		if( tb[IFA_LABEL] )
		{
		    strncpy
			( &(nla->label[0]), 
			  nla_data((void*)tb[IFA_LABEL]), 
			  nla_len((void*)tb[IFA_LABEL])
			);
		    eprintf(NIC_DEBUG,"nic %d addr label: %s", 
			    nla->ifa.ifa_index, nla->label);
		    nla->have |= NIC_ADDR_LABEL;
		}
		
		n_addrs++;
		
		struct nic_ip_address_s **paddr, *taddr=0L, *naddr = 0L;

		if( ipa == 0 )
		{
		    naddr = calloc(1, sizeof(struct nic_ip_address_s));
		    if( naddr == 0L )
		    {
			eprintf(NIC_ERR, "nic_get_addresses: out of memory.");
			goto nla_return;
		    }
		    *naddr = *addr;
		    taddr = addr;
		    addr = naddr;
		    naddr = taddr;
		    taddr = 0;
		}
		
		if(  (nh->nic_addr_tree == 0L)
		  ||((paddr = tfind(addr, &(nh->nic_addr_tree), nic_addr_comparator)) == 0L)
		  ) tsearch(addr, &(nh->nic_addr_tree), nic_addr_comparator);
		
		if( paddr != 0L )
		{
		    taddr = *paddr;
		    *taddr = *addr;
		    if( naddr )
			free(naddr);
		}

		if( ipa == 0L )
		    addr = &ipaf;

		if( ipa && ( nic_addr_comparator( ipa , addr ) == 0 ) )
		    goto nla_return;

		hdr = nlmsg_next( hdr, &remaining );

	    } while ( remaining > 0 );	

	}else
	{
            eprintf(NIC_ERR,"nic_addresses: netlink recv failed: %s %d\n", 
		    nl_get_errno(), nl_geterror());
	    goto nla_return;
	}
    } while ( (hdr->nlmsg_type == RTM_NEWADDR) && (remaining > 0) );

nla_return:
    if( buf )
	free(buf);
    nlmsg_free(msg);
    return n_addrs;	    

nla_put_failure:
    nlmsg_free(msg);
    return n_addrs;
}

struct nic_addr_cbarg
{
    IPaddr_Handler_t cb;
    void        *arg;
};

static void nic_addr_twalker( const void *p, const VISIT which, const int depth )
{
    struct nic_ip_address_s *addr, *const*app=p;
    struct nic_addr_cbarg *cb;

    if( (app == 0L)    || ((addr = *app) == 0L)
      ||(addr->nh == 0) || (addr->nh->nic_addr_foreach_arg == 0 )
      ||( (which != postorder) && (which != leaf) )
      ) return;

    cb = addr->nh->nic_addr_foreach_arg;
 
    cb->cb( addr->nh, addr, cb->arg );       
}

void nic_addr_foreach(NLH_t nh, IPaddr_Handler_t handler,  void *arg)
{
    struct nic_addr_cbarg cb = { handler, arg };

    if ( nh->nic_addr_foreach_arg )
	return;

    nh->nic_addr_foreach_arg = &cb;

    nic_get_addresses(nh, 0L);

    if( nh->nic_addr_tree )
	twalk( nh->nic_addr_tree, nic_addr_twalker);

    nh->nic_addr_foreach_arg = 0L;
}

IPaddr_t nic_addr_ip( NLH_t nh, ip_addr_t *ip)
{
    if( ( nh == 0L ) || ( ip == 0L ) 
      ||( (ip->sa_family != AF_INET)
        &&(ip->sa_family != AF_INET6)
	)
      )	return 0;
	
    IPaddr_t ipa = calloc(1, sizeof(struct nic_ip_address_s));
    
    if( ipa == 0L )
    {
        eprintf(NIC_ERR, "nic_addr_ip: out of memory");
    }
    ipa->nh = nh;
    ipa->nla.addr = *ip;
    ipa->nla.have |= NIC_ADDR_ADDRESS;
    return ipa;
}

IPaddr_t nic_addr( NLH_t nh, ip_addr_t ip)
{
    if( ( nh == 0L )
      ||( (ip.sa_family != AF_INET)
        &&(ip.sa_family != AF_INET6)
	)
      )	return 0;
	
    IPaddr_t ipa = calloc(1, sizeof(struct nic_ip_address_s));
    
    if( ipa == 0L )
    {
        eprintf(NIC_ERR, "nic_addr: out of memory");
    }
    ipa->nh = nh;
    ipa->nla.addr = ip;
    ipa->nla.ifa.ifa_family = ip.sa_family;
    ipa->nla.have |= NIC_ADDR_ADDRESS;
    return ipa;
}

IPaddr_t nic_addr_local( NLH_t nh, ip_addr_t ip)
{
    if( ( nh == 0L )
      ||((ip.sa_family != AF_INET)
       &&(ip.sa_family != AF_INET6)
	)
      )	return 0;
	
    IPaddr_t ipa = calloc(1, sizeof(struct nic_ip_address_s));
    
    if( ipa == 0L )
    {
        eprintf(NIC_ERR, "nic_addr_local: out of memory");
    }
    ipa->nh = nh;
    ipa->nla.local = ip;
    ipa->nla.have |= NIC_ADDR_LOCAL;
    return ipa;
}

ip_addr_t nic_ip_addr( IPaddr_t ipa )
{
    ip_addr_t ip;
    memset(&ip, '\0', sizeof(ip_addr_t));
    if( (ipa == 0L) || ((ipa->nla.have & NIC_ADDR_ADDRESS) == 0))
	return ip;
    return ipa->nla.addr;
}

ip_addr_t nic_addr_get_local( IPaddr_t ipa )
{
    ip_addr_t ip;
    memset(&ip, '\0', sizeof(ip_addr_t));
    if( (ipa == 0L) || ((ipa->nla.have & NIC_ADDR_LOCAL) == 0))
	return ip;
    return ipa->nla.local;
}

void nic_addr_set_local( IPaddr_t ipa, ip_addr_t ip )
{
    if((ipa == 0L) 
      ||((ip.sa_family != AF_INET)
       &&(ip.sa_family != AF_INET6)
	)
      )	return;
    
    ipa->nla.local = ip;
    ipa->nla.have |= NIC_ADDR_LOCAL;
}

uint8_t nic_addr_get_family( IPaddr_t ipa )
{
    if(ipa == 0L) return 0;
    return ipa->nla.ifa.ifa_family;
}

uint8_t nic_addr_get_prefix( IPaddr_t ipa)
{
    if(ipa == 0L) return 0;
    return ipa->nla.ifa.ifa_prefixlen;
}

void nic_addr_set_prefix( IPaddr_t ipa, uint8_t prefix)
{
    if(  (ipa == 0L) 	 
       ||((ipa->nla.ifa.ifa_family == AF_INET) && (prefix > 32))
       ||((ipa->nla.ifa.ifa_family == AF_INET6) && (prefix > 128))
      )
	return;
    
    ipa->nla.ifa.ifa_prefixlen = prefix;    
}

ip_addr_t nic_addr_get_broadcast( IPaddr_t ipa )
{
    ip_addr_t ip;
    memset(&ip, '\0', sizeof(ip_addr_t));
    if( (ipa == 0L) || ((ipa->nla.have & NIC_ADDR_BROADCAST) == 0))
	return ip;
    return ipa->nla.broadcast;
}

void nic_addr_set_broadcast( IPaddr_t ipa, ip_addr_t ip )
{
    if( (ipa == 0L) 
      ||((ip.sa_family != AF_INET)
       &&(ip.sa_family != AF_INET6)
        )
      )	return;
    
    ipa->nla.broadcast = ip;
    ipa->nla.have |= NIC_ADDR_BROADCAST;
}

ip_addr_t nic_addr_get_anycast( IPaddr_t ipa )
{
    ip_addr_t ip;
    memset(&ip, '\0', sizeof(ip_addr_t));
    if( (ipa == 0L) || ((ipa->nla.have & NIC_ADDR_ANYCAST) == 0))
	return ip;
    return ipa->nla.anycast;
}

void nic_addr_set_anycast( IPaddr_t ipa, ip_addr_t ip )
{
    if( (ipa == 0L) 
      ||((ip.sa_family != AF_INET)
       &&(ip.sa_family != AF_INET6)
	)
      )	return;
    
    ipa->nla.anycast = ip;
    ipa->nla.have |= NIC_ADDR_ANYCAST;
}

ip_addr_t nic_addr_get_multicast( IPaddr_t ipa )
{
    ip_addr_t ip;
    memset(&ip, '\0', sizeof(ip_addr_t));
    if( (ipa == 0L) || ((ipa->nla.have & NIC_ADDR_MULTICAST) == 0))
	return ip;
    return ipa->nla.multicast;
}

void nic_addr_set_multicast( IPaddr_t ipa, ip_addr_t ip )
{
    if( (ipa == 0L) 
      ||( (ip.sa_family != AF_INET)
        &&(ip.sa_family != AF_INET6)
	)
      )	return;
    
    ipa->nla.multicast = ip;
    ipa->nla.have |= NIC_ADDR_MULTICAST;
}

int8_t nic_addr_get_scope(IPaddr_t ipa)
{
    if(ipa == 0L) return 0;

    return ipa->nla.ifa.ifa_scope;	
}

void nic_addr_set_scope(IPaddr_t ipa, int8_t scope )
{
    if(ipa == 0L) return;
    
    ipa->nla.ifa.ifa_scope = scope;
}

uint8_t nic_addr_get_flags(IPaddr_t ipa)
{
    if(ipa == 0L) return 0;

    return ipa->nla.ifa.ifa_flags;	
}

void nic_addr_set_flags(IPaddr_t ipa, uint8_t flags )
{
    if(ipa == 0L) return;
    
    ipa->nla.ifa.ifa_flags = flags;
}

const char *nic_addr_get_label(IPaddr_t ipa)
{
    if( (ipa == 0L) || ((ipa->nla.have & NIC_ADDR_LABEL) == 0))
	return "";
    return &(ipa->nla.label[0]);       
}

void nic_addr_set_label(IPaddr_t ipa, const char *label)
{
    if((ipa == 0L) || (label == 0L)) return;
    strncpy(&(ipa->nla.label[0]), label, IFNAMSIZ);
    ipa->nla.have |= NIC_ADDR_LABEL;
}

struct ifa_cacheinfo nic_addr_get_cacheinfo( IPaddr_t ipa )
{
    struct ifa_cacheinfo ifa_ca;
    memset(&ifa_ca, '\0', sizeof(struct ifa_cacheinfo));
    if( (ipa == 0L) || ((ipa->nla.have & NIC_ADDR_CACHEINFO) == 0))
	return ifa_ca;
    return ipa->nla.cacheinfo;
}

void  nic_addr_set_cacheinfo( IPaddr_t ipa, struct ifa_cacheinfo *ca)
{
    if((ipa == 0L) || (ca == 0L)) return;
    ipa->nla.cacheinfo = *ca;
    ipa->nla.have |= NIC_ADDR_CACHEINFO;
}

IPaddr_list_t *nic_address_list_new( IPaddr_t ip1, ... )
{      
    va_list va;
    NLH_t nh = ip1->nh;
    va_start(va, ip1);
    
    IPaddr_list_t *head = calloc(1,sizeof(IPaddr_list_t));
    IPaddr_list_node_t *n;
    struct nic_ip_address_s *ip;

    if( head == 0 )
    {
	va_end(va);
        eprintf(NIC_ERR, "out of memory" );
	return 0;
    }
    ip=ip1;
    STAILQ_INIT(head);
    do
    {
	n = calloc(1,sizeof(IPaddr_list_node_t));
	if ( n == 0L )
	{
	    va_end(va);
	    free(head);
            eprintf(NIC_ERR, "out of memory" );
	    return 0L;
	}
	n->addr = ip;
	STAILQ_INSERT_TAIL( head, n, link );
    }while( (ip = va_arg(va, struct nic_ip_address_s*)) != 0) ;
    va_end(va);
    return(head);
}

void nic_address_list_free( IPaddr_list_t *head )
{
    IPaddr_list_node_t *n, *a;
    
    if ( !STAILQ_EMPTY(head) )
    {
	n = STAILQ_FIRST( head );
	do
	{
	    a = STAILQ_NEXT(n, link);
	    free(n->addr);
	    free(n);
	    n = a;
	} while (n);
    }
}

typedef
enum nic_ar_e
{
    NIC_ADD_ADDRESS,
    NIC_DELETE_ADDRESS
} NIC_Address_Request;

static 
int nic_addr_send( IPaddr_t ipa, NIC_Address_Request req )
{
    char abuf[64];
    int err=0;
    NLH_t nh;

    if(ipa == 0L)
        return -1;
    nh = ipa->nh;
    
    memset(&(ipa->nla.hdr), '\0', sizeof(struct nlmsghdr));

    ipa->nla.hdr.nlmsg_type = 
	(req == NIC_DELETE_ADDRESS) 
	? RTM_DELADDR
	: RTM_NEWADDR;

    ipa->nla.hdr.nlmsg_flags =
	( (req == NIC_ADD_ADDRESS) 
	  ? (NLM_F_REPLACE | NLM_F_MATCH)
	  : 0
        ) | NLM_F_REQUEST | NLM_F_ACK |  NLM_F_ROOT;

    ipa->nla.hdr.nlmsg_len = sizeof(struct nlmsghdr);

    if( ipa->nla.ifa.ifa_flags == 0 )
	ipa->nla.ifa.ifa_flags = IFA_F_PERMANENT;

    struct nl_msg *msg = nlmsg_build(&(ipa->nla.hdr));

    if ( msg == 0L ) {
	eprintf(NIC_ERR, "nic_addr_send: nlmsg_build failed - %d %s" ,
		nl_get_errno(), nl_geterror());
	return NIC_FAIL;
    }

    nlmsg_append( msg, &(ipa->nla.ifa), sizeof( struct ifaddrmsg ), 1);

    if( ipa->nla.have & NIC_ADDR_ADDRESS )
    {
	NLA_PUT(msg, IFA_ADDRESS, IP_ADDR_SIZE(&(ipa->nla.addr)), IP_ADDR( &(ipa->nla.addr) ));
	if( (ipa->nla.ifa.ifa_family == AF_INET)  && ((ipa->nla.have & NIC_ADDR_LOCAL)==0) )
	{
	    ipa->nla.have |= NIC_ADDR_LOCAL;
	    ipa->nla.local = ipa->nla.addr;
	}
        eprintf(NIC_DEBUG,"nic_addr_send: add IFA_ADDRESS: %d %s/%d\n", 
		ipa->nla.ifa.ifa_index, ip_text_addr(&(ipa->nla.addr),abuf,64),
		ipa->nla.ifa.ifa_prefixlen);
    }

    if( ipa->nla.have & NIC_ADDR_LOCAL )
    {
	NLA_PUT(msg, IFA_LOCAL, IP_ADDR_SIZE(&(ipa->nla.local)), IP_ADDR( &(ipa->nla.local) ));
        eprintf(NIC_DEBUG,"nic_addr_send: add IFA_LOCAL: %d %s\n",
		ipa->nla.ifa.ifa_index,
                ip_text_addr(&(ipa->nla.local),abuf,64));
    }

    if( ipa->nla.have & NIC_ADDR_LABEL )
	NLA_PUT_STRING(msg, IFA_LABEL, &(ipa->nla.label[0]));

    if( ipa->nla.have & NIC_ADDR_BROADCAST )
    {
	NLA_PUT(msg, IFA_BROADCAST, IP_ADDR_SIZE(&(ipa->nla.broadcast)), IP_ADDR( &(ipa->nla.broadcast) ));
	eprintf(NIC_DEBUG,"nic_addr_send: add address: %d broadcast: %s\n", 
		ipa->nla.ifa.ifa_index,
                ip_text_addr(&(ipa->nla.broadcast), abuf, 64));
    }

    if( ipa->nla.have & NIC_ADDR_ANYCAST )
	NLA_PUT(msg, IFA_ANYCAST, IP_ADDR_SIZE(&(ipa->nla.broadcast)), IP_ADDR( &(ipa->nla.broadcast) ));
    
    if( ipa->nla.have & NIC_ADDR_MULTICAST )
	NLA_PUT(msg, IFA_MULTICAST, IP_ADDR_SIZE(&(ipa->nla.broadcast)), IP_ADDR( &(ipa->nla.multicast) ));

    if( ipa->nla.have & NIC_ADDR_CACHEINFO )
	NLA_PUT(msg, IFA_CACHEINFO, sizeof(struct ifa_cacheinfo), &(ipa->nla.cacheinfo));

    if( (err = nl_send_auto_complete(ipa->nh->nl, nlmsg_hdr(msg))) < 0 )
    {
        eprintf(NIC_ERR,"nic_addr_send: send failed - %d %d %s",
		err, nl_get_errno(), nl_geterror());
	return NIC_FAIL;
    }

    nlmsg_free(msg);

    int n_tries = 0;
    try_again:

    if( (n_tries < 10) && ((err= nl_wait_for_ack(ipa->nh->nl)) < 0) )
    {
	if( err == - EBUSY )
	{
	    struct timespec ts = { 0, 500000000 };
	    ++n_tries;
	    /* Device or resource busy */
	    nanosleep(&ts,0);
	    goto try_again;
	}
	
	if( err == -EEXIST )
	{
	    /* something isn't listening to our "NLM_F_REPLACE" flag ...*/
	    return NIC_OK;
	}    
        eprintf(NIC_ERR,"nic_addr_send failed - %d: %s", nl_get_errno(),
                nl_geterror());
	return NIC_FAIL;	
    }
    
    return NIC_OK;

nla_put_failure:

    eprintf(NIC_ERR, "nic_addr_send append attr failed - out of memory?");

    nlmsg_free(msg);

    return NIC_FAIL;
}

NIC_Res_t nic_add_address( NIC_t nic, IPaddr_t ipa)
{
    if( ( ipa == 0L ) || (nic == 0L) )
	return NIC_FAIL;
    ipa->nla.ifa.ifa_index = nic->l.ifi.ifi_index;
    return nic_addr_send( ipa, NIC_ADD_ADDRESS );
}

NIC_Res_t nic_remove_address( NIC_t nic, IPaddr_t ipa)
{
    if( ( ipa == 0L ) || (nic == 0L) )
	return NIC_FAIL;
    ipa->nla.ifa.ifa_index = nic->l.ifi.ifi_index;
    return nic_addr_send( ipa, NIC_ADD_ADDRESS );
}

NIC_Res_t nic_add_addresses( NIC_t nic, IPaddr_list_t *head )
{
    IPaddr_list_node_t *n;
    NIC_Res_t r;

    STAILQ_FOREACH(n,head,link)
    {
	if((r = nic_add_address(nic, n->addr)) == NIC_FAIL)
	    return r;
    }
    return NIC_OK;
}

NIC_Res_t nic_remove_addresses( NIC_t nic, IPaddr_list_t *head )
{
    IPaddr_list_node_t *n;
    NIC_Res_t r;

    STAILQ_FOREACH(n,head,link)
    {
	if((r = nic_remove_address(nic, n->addr)) == NIC_FAIL)
	    return r;
    }
    return NIC_OK;
}

void nic_addr_free(void *ipa)
{
    if (ipa)
        free(ipa);
}

struct nic_route_s
{
    NLH_t nh;
    
    struct nic_rt_s
    {
	struct nlmsghdr hdr;
	struct rtmsg    rtm;
	/* struct rtmsg
	 * {
	 *	unsigned char		rtm_family;
	 *	unsigned char		rtm_dst_len;
	 *	unsigned char		rtm_src_len;
	 *	unsigned char		rtm_tos;
	 *
	 *	unsigned char		rtm_table;
	 *	unsigned char		rtm_protocol;
	 *	unsigned char		rtm_scope;
	 *	unsigned char		rtm_type;
	 *
	 *	unsigned		rtm_flags;
	 *};
	 */

	ip_addr_t dst;
	ip_addr_t gw;
	ip_addr_t src;
	ip_addr_t prefsrc;
	int32_t   oif;       
	uint32_t  priority;
	uint32_t  protoinfo;
	uint32_t  flow;
	uint32_t  session;
	uint32_t  mp_algo;
	struct rta_cacheinfo cacheinfo;
	char      iif [ IFNAMSIZ ];
	uint32_t  metrics[ __RTAX_MAX ];

	struct nic_rtnh_s
	{
	    struct rtnexthop rtnh;
	    ip_addr_t gw;
	} *hops;
	uint32_t n_hops;
	
	enum rt_a_e
	{
	    NIC_RT_DST      =   1,
	    NIC_RT_SRC      =   2,  
	    NIC_RT_IIF      =   8,
	    NIC_RT_OIF      =  16,
	    NIC_RT_GATEWAY  =  32,
	    NIC_RT_PRIORITY =  64,
	    NIC_RT_PREFSRC  = 128,
	    NIC_RT_METRICS  = 256,
	    NIC_RT_MULTIPATH= 512,
	    NIC_RT_PROTOINFO=1024,
	    NIC_RT_FLOW     =2048,
	    NIC_RT_CACHEINFO=4096,
	    NIC_RT_SESSION  =8192,
	    NIC_RT_MP_ALGO= 16384
	} have, change;
	
	enum rtax_e
	{
	    NIC_RTAX_LOCK =	(1<<RTAX_LOCK),
	    NIC_RTAX_MTU  =     (1<<RTAX_MTU),
	    NIC_RTAX_WINDOW =   (1<<RTAX_WINDOW),
	    NIC_RTAX_RTT    =   (1<<RTAX_RTT),
	    NIC_RTAX_RTTVAR =   (1<<RTAX_RTTVAR),
	    NIC_RTAX_SSTHRESH = (1<<RTAX_SSTHRESH),
	    NIC_RTAX_CWND     = (1<<RTAX_CWND),
	    NIC_RTAX_ADVMSS   = (1<<RTAX_ADVMSS),
	    NIC_RTAX_REORDERING=(1<<RTAX_REORDERING),
	    NIC_RTAX_HOPLIMIT  =(1<<RTAX_HOPLIMIT),
	    NIC_RTAX_INITCWND  =(1<<RTAX_INITCWND),
	    NIC_RTAX_FEATURES  =(1<<RTAX_FEATURES)
	} have_metrics, change_metrics;
    } rt;
};

typedef struct nic_route_s RT_t;

static struct nla_policy nic_route_policy[RTA_MAX+1] = {
	[RTA_IIF]	= { .type = NLA_STRING,
			    .maxlen = IFNAMSIZ, },
	[RTA_OIF]	= { .type = NLA_U32 },
	[RTA_PRIORITY]	= { .type = NLA_U32 },
	[RTA_FLOW]	= { .type = NLA_U32 },
	[RTA_MP_ALGO]	= { .type = NLA_U32 },
	[RTA_CACHEINFO]	= { .minlen = sizeof(struct rta_cacheinfo) },
	[RTA_METRICS]	= { .type = NLA_NESTED },
	[RTA_MULTIPATH]	= { .type = NLA_NESTED },
};

static int nic_route_comparator( const void *p1, const void *p2 )
{
    const RT_t *rt1 = p1, *rt2 = p2;

    if( rt1->rt.rtm.rtm_table != rt2->rt.rtm.rtm_table )
	return
	(   (rt1->rt.rtm.rtm_table > rt2->rt.rtm.rtm_table )
	    ? 1
	    : -1
        );

    if ( rt1->rt.rtm.rtm_family != rt2->rt.rtm.rtm_family )
	return
	(   ( rt1->rt.rtm.rtm_family > rt2->rt.rtm.rtm_family )
	    ? 1
	    : -1
	);
 
    if ( rt1->rt.rtm.rtm_type != rt2->rt.rtm.rtm_type )
	return
	(   ( rt1->rt.rtm.rtm_type > rt2->rt.rtm.rtm_type )
	    ? 1
	    : -1
	);

    if ( rt1->rt.rtm.rtm_scope != rt2->rt.rtm.rtm_scope )
	return
	(   ( rt1->rt.rtm.rtm_scope > rt2->rt.rtm.rtm_scope )
	    ? 1
	    : -1
	);

    if ( rt1->rt.rtm.rtm_protocol != rt2->rt.rtm.rtm_protocol )
	return
	(   ( rt1->rt.rtm.rtm_protocol > rt2->rt.rtm.rtm_protocol )
	    ? 1
	    : -1
	);

    if( rt1->rt.have & NIC_RT_OIF )
    {	
	if( ( rt2->rt.have & NIC_RT_OIF ) == 0 )
	    return 1;
	else
	{
	    if( rt1->rt.oif != rt2->rt.oif )
		return
		    (   (rt1->rt.oif > rt2->rt.oif)
			? 1
			: -1
		    ); 
	}
    }else
    if( rt2->rt.have & NIC_RT_OIF )
	return -1;

    if( rt1->rt.have & NIC_RT_IIF )
    {	
	if( ( rt2->rt.have & NIC_RT_IIF ) == 0 )
	    return 1;
	else
	{
	    int c = strcmp(&(rt1->rt.iif[0]),&(rt2->rt.iif[0]));
	    if( c != 0 )
		return c;
	}
    }else
    if( rt2->rt.have & NIC_RT_OIF )
	return -1;

    if( rt1->rt.have & NIC_RT_DST )
    {
	if( (rt2->rt.have & NIC_RT_DST) == 0 )
	    return 1;
	else
	{
		
	    if( rt1->rt.dst.sa_family != rt2->rt.dst.sa_family )
		return
		(   ( rt1->rt.dst.sa_family > rt2->rt.dst.sa_family )
		    ? 1
		    : -1
		);

	    if( rt1->rt.rtm.rtm_dst_len != rt2->rt.rtm.rtm_dst_len )
		return
	        (   ( rt1->rt.rtm.rtm_dst_len > rt2->rt.rtm.rtm_dst_len )
		    ? 1
		    : -1
		);

	    uint8_t
		*ap1 = IP_ADDR( &(rt1->rt.dst) ),
		*ap2 = IP_ADDR( &(rt2->rt.dst) ),
		pfx1 = rt1->rt.rtm.rtm_dst_len;

	    if( pfx1 == 0 )
		pfx1 = IP_ADDR_SIZE( &(rt1->rt.dst) );

	    int cr = memcmp( ap1, ap2,
			     pfx1 / 8
		           );
	    if ( cr != 0 )
		return cr;
	    uint8_t 
		v1 = ((uint8_t*)IP_ADDR( &(rt1->rt.dst) ))[ pfx1 / 8 ]
		    & ( ( 1 << ( pfx1 % 8 ) ) - 1),
	        v2 = ((uint8_t*)IP_ADDR( &(rt2->rt.dst) ))[ pfx1 / 8 ]
		    & ( ( 1 << ( pfx1 % 8 ) ) - 1);		

	    if( v1 != v2 )
		return 
	        (
		    (v1 > v2)
		    ?  1
		    : -1
		);
	    
	    if( rt1->rt.have & NIC_RT_PRIORITY )
	    {
		if( (rt2->rt.have & NIC_RT_PRIORITY) == 0 )
		    return -1;
		
		if( rt1->rt.priority !=  rt2->rt.priority )
		    return
		    (   (rt1->rt.priority >  rt2->rt.priority)
			? -1
                        : 1
		    );
	    }
	}
    }else
    if( rt2->rt.have & NIC_RT_DST )
	return -1;

    if( rt1->rt.have & NIC_RT_SRC )
    {
	if( (rt2->rt.have & NIC_RT_SRC) == 0 )
	    return 1;
	else
	{		
	    if( rt1->rt.src.sa_family != rt2->rt.src.sa_family )
		return
		(   ( rt1->rt.src.sa_family > rt2->rt.src.sa_family )
		    ? 1
		    : -1
		);

	    if( rt1->rt.rtm.rtm_src_len != rt2->rt.rtm.rtm_src_len )
		return
	        (   ( rt1->rt.rtm.rtm_src_len > rt2->rt.rtm.rtm_src_len )
		    ? 1
		    : -1
		);

	    uint8_t
		*ap1 = IP_ADDR( &(rt1->rt.src) ),
		*ap2 = IP_ADDR( &(rt2->rt.src) ),
		pfx1 = rt1->rt.rtm.rtm_src_len;

	    if( pfx1 == 0 )
		pfx1 = IP_ADDR_SIZE( &(rt1->rt.src) );

	    int cr = memcmp( ap1, ap2,
			     rt1->rt.rtm.rtm_src_len / 8
		           );
	    if ( cr != 0 )
		return cr;
	    uint8_t 
		v1 = ((uint8_t*)IP_ADDR( &(rt1->rt.src) ))[ pfx1 / 8 ]
		    & ( ( 1 << ( rt1->rt.rtm.rtm_src_len % 8 ) ) - 1),
	        v2 = ((uint8_t*)IP_ADDR( &(rt2->rt.src) ))[ pfx1 / 8 ]
		    & ( ( 1 << ( pfx1 % 8 ) ) - 1);		

	    if( v1 != v2 )
		return 
	        (
		    (v1 > v2)
		    ?  1
		    : -1
		);
	    
	    if( rt1->rt.have & NIC_RT_PRIORITY )
	    {
		if( (rt2->rt.have & NIC_RT_PRIORITY) == 0 )
		    return -1;
		
		if( rt1->rt.priority !=  rt2->rt.priority )
		    return
		    (   (rt1->rt.priority >  rt2->rt.priority)
			? 1
                        : -1
		    );
	    }
	}
    }else
    if (rt2->rt.have & NIC_RT_SRC)
	return -1;

    if ( rt1->rt.have & NIC_RT_GATEWAY )
    {
	if((rt2->rt.have & NIC_RT_GATEWAY)==0)
	    return 1;
	{

	    if( rt1->rt.gw.sa_family != rt2->rt.gw.sa_family )
		return
		(   ( rt1->rt.gw.sa_family > rt2->rt.gw.sa_family )
		    ? 1
		    : -1
		);
		    
	    uint8_t
		*ap1 = IP_ADDR( &(rt1->rt.gw) ),
		*ap2 = IP_ADDR( &(rt2->rt.gw) );

	    int cr = memcmp( ap1, ap2,
			     IP_ADDR_SIZE(&(rt1->rt.gw))
		           );
	    
	    if( cr != 0 )
		return cr;	    
	}
    }else
    if (rt2->rt.have & NIC_RT_GATEWAY)
	return -1;

    if ( rt1->rt.have & NIC_RT_PREFSRC )
    {
	if((rt2->rt.have & NIC_RT_PREFSRC)==0)
	    return 1;
	{

	    if( rt1->rt.prefsrc.sa_family != rt2->rt.prefsrc.sa_family )
		return
		(   ( rt1->rt.prefsrc.sa_family > rt2->rt.prefsrc.sa_family )
		    ? 1
		    : -1
		);
		    
	    uint8_t
		*ap1 = IP_ADDR( &(rt1->rt.prefsrc) ),
		*ap2 = IP_ADDR( &(rt2->rt.prefsrc) );

	    int cr = memcmp( ap1, ap2,
			     IP_ADDR_SIZE(&(rt1->rt.prefsrc))
		           );
	    
	    if( cr != 0 )
		return cr;	    
	}
    } else
    if (rt2->rt.have & NIC_RT_PREFSRC)
	return -1;
    return 0;
}

static int nic_get_routes(NLH_t nh, RT_t *rt)
{
    RT_t srt, *rtp=0;
    struct nic_rt_s  *rta=0;
    int n_routes = 0;

    if(nh == 0)
	return 0;   

    if( rt )
    {
	srt = *rt;
	rtp = rt;
	rta = &(rt->rt);    
	memset(&(rta->hdr),'\0',sizeof(struct nlmsghdr));
	rta->hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_ATOMIC | NLM_F_MATCH;
    }else
    {
	memset(&srt, '\0', sizeof(RT_t));
	srt.nh = nh;
	rtp = &srt;
	rta = &(rtp->rt);    
	rta->hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
    }
    
    rta->hdr.nlmsg_type = RTM_GETROUTE;
    
    struct nl_msg *msg = nlmsg_build(&(rta->hdr));

    if( msg == 0L )
    {
	eprintf(NIC_ERR, "nic_get_routes: out of memory" );
	return n_routes;
    }
    
    if( nlmsg_append(msg, &(rta->rtm), sizeof(struct rtmsg), 1) < 0 )
    {	
        eprintf(NIC_ERR,"nic_get_routes: nlmsg_append failed: %d %s\n", 
		nl_get_errno(), nl_geterror());
	goto nla_put_failure;
    }

    if(  rta->rtm.rtm_dst_len && rta->dst.sa_family 
      &&(rta->have & NIC_RT_DST)
      )
	NLA_PUT(msg, RTA_DST, 
		IP_ADDR_SIZE(&(rta->dst)), 
		IP_ADDR(&(rta->dst))
	       );

    if( (rta->have & NIC_RT_OIF) &&  (rta->oif > -1) )
	NLA_PUT_U32(msg, RTA_OIF, rta->oif);
    
    if(  rta->gw.sa_family 
      &&(rta->have & NIC_RT_GATEWAY)
      )
	NLA_PUT(msg, RTA_GATEWAY, 
		IP_ADDR_SIZE(&(rta->gw)), 
		IP_ADDR(&(rta->gw))
	       );
    
    if(  rta->rtm.rtm_src_len && rta->src.sa_family 
      &&(rta->have & NIC_RT_SRC)
      )
	NLA_PUT(msg, RTA_SRC, 
		IP_ADDR_SIZE(&(rta->src)), 
		IP_ADDR(&(rta->src))
	       );
    
    if( (rta->have & NIC_RT_IIF) &&  (rta->iif[0] != '\0') )
	NLA_PUT_STRING(msg, RTA_IIF, &(rta->iif[0]));    	

    if ( nl_send_auto_complete( nh->nl, nlmsg_hdr(msg) ) < 0 )
    {
        eprintf(NIC_ERR,"nic_get_routes: send failed: %d %s\n", 
		nl_get_errno(), nl_geterror());
	goto nla_put_failure;
    }

    char abuf[64];
    struct RT_t *buf =0;
    struct nlmsghdr *hdr = (struct nlmsghdr *)buf;
    struct sockaddr_nl kernel;
    int rlen = 0;
    int remaining = 0;
    
    do
    {
	if ( (rlen = nl_recv( nh->nl, &kernel, (void*)&buf)) > 0 )
	{
	    if( rlen < (sizeof(struct nlmsghdr) + sizeof(struct rtmsg)) )
		goto rta_return;

	    hdr = (struct nlmsghdr*)buf;
	    remaining = rlen;

	    if( hdr->nlmsg_type != RTM_NEWROUTE )
		goto rta_return;
	    
	    do
	    {
		memset(rta, '\0', sizeof(struct nic_rt_s));
		
		memcpy(rta, hdr, sizeof(struct nlmsghdr));
		
		struct nl_attr *tb[__RTA_MAX];		
		int i;
		for (i=0; i < __RTA_MAX; i++) tb[i]=0;

		int err = 
		    nlmsg_parse((struct nlmsghdr*)hdr, sizeof(struct rtmsg),
				(void*)tb, RTA_MAX, nic_route_policy);
		if ( err < 0 )
		{
                    eprintf(NIC_ERR,"nic_get_routes: nl_parse failed: %d %s",
			    err, nl_geterror());
		    goto rta_return;
		}
		
		rta->rtm = *((struct rtmsg*)nlmsg_data((struct nlmsghdr*)hdr));

		eprintf(NIC_DEBUG, "\n"
                                   "table: %d\ndst_len: %d\nsrc_len: %d\n"
                                   "scope: %d\ntype: %d\nprotocol: %d", 
			rta->rtm.rtm_table, rta->rtm.rtm_dst_len,
                        rta->rtm.rtm_src_len, rta->rtm.rtm_scope,
                        rta->rtm.rtm_type,  rta->rtm.rtm_protocol);

		if ( tb[RTA_DST] )
		{
		    rta->dst =
		    ip_addr_binary 
			( nla_data((void*)tb[RTA_DST]),
			  nla_len((void*)tb[RTA_DST])
			);
		    rta->have |= NIC_RT_DST;
                    eprintf(NIC_DEBUG, "nic route: dst: %s/%d",
			    ip_text_addr(&(rta->dst),abuf,64),
			    rta->rtm.rtm_dst_len);
		}
		
		if ( tb[RTA_OIF] )
		{
		    rta->oif = nla_get_u32( (void*) tb[RTA_OIF] );
		    rta->have |= NIC_RT_OIF ;
                    eprintf(NIC_DEBUG, "nic route: oif: %d", rta->oif);			
		}

		if ( tb[RTA_GATEWAY] )
		{
		    rta->gw =
		    ip_addr_binary 
			( nla_data((void*)tb[RTA_GATEWAY]),
			  nla_len((void*)tb[RTA_GATEWAY])
			);
		    rta->have |= NIC_RT_GATEWAY;
		    eprintf(NIC_DEBUG, "nic route: gw: %s",
			    ip_text_addr(&(rta->gw),abuf,64));
		}

		if ( tb[RTA_SRC] )
		{
		    rta->src =
		    ip_addr_binary 
			( nla_data((void*)tb[RTA_SRC]),
			  nla_len((void*)tb[RTA_SRC])
			);
		    rta->have |= NIC_RT_SRC;
		    eprintf(NIC_DEBUG, "nic route: src: %s/%d",
			    ip_text_addr(&(rta->src),abuf,64),
			    rta->rtm.rtm_src_len);
		}
		
		if ( tb[RTA_IIF] )
		{
		    memcpy( &(rta->iif),
			    nla_data( (void*) tb[RTA_IIF] ),
			    nla_len((void*) tb[RTA_IIF] )
			  );
		    rta->have |= NIC_RT_IIF ;
                    eprintf(NIC_DEBUG, "nic route: iif: %s", rta->iif);			
		}
		
		if ( tb[RTA_PREFSRC] )
		{
		    rta->prefsrc =
		    ip_addr_binary 
			( nla_data((void*)tb[RTA_PREFSRC]),
			  nla_len((void*)tb[RTA_PREFSRC])
			);
		    rta->have |= NIC_RT_PREFSRC;
		    eprintf(NIC_DEBUG, "nic route: prefsrc: %s",
			    ip_text_addr(&(rta->prefsrc),abuf,64));
		}

		if ( tb[RTA_PRIORITY] )
		{
		    rta->priority = nla_get_u32( (void*)tb[RTA_PRIORITY]);
		    rta->have |= NIC_RT_PRIORITY;
                    eprintf(NIC_DEBUG,"nic route: priority: %d", rta->priority);
		}
		
		if ( tb[RTA_PROTOINFO] )
		{
		    rta->protoinfo = nla_get_u32( (void*)tb[RTA_PROTOINFO]);
		    rta->have |= NIC_RT_PROTOINFO;
                    eprintf(NIC_DEBUG, "nic route: protoinfo: %x", rta->protoinfo);		    
		}

		if ( tb[RTA_FLOW] )
		{
		    rta->flow = nla_get_u32( (void*)tb[RTA_FLOW]);
		    rta->have |= NIC_RT_FLOW;
                    eprintf(NIC_DEBUG, "nic route: flow: %x", rta->flow); 
		}

		if ( tb[RTA_CACHEINFO] )
		{
		    memcpy( &(rta->cacheinfo),
			    nla_data((void*)tb[RTA_CACHEINFO]),
			    sizeof(struct rta_cacheinfo)
			  );
		    rta->have |= NIC_RT_CACHEINFO;
                    eprintf(NIC_DEBUG, "nic route: have cacheinfo");
		}

		if ( tb[RTA_SESSION] )
		{
		    rta->session = nla_get_u32( (void*)tb[RTA_SESSION]);
		    rta->have |= NIC_RT_SESSION;
                    eprintf(NIC_DEBUG, "nic route: session: %d", rta->session);
		}
		
		if ( tb[RTA_MP_ALGO] )
		{
		    rta->mp_algo = nla_get_u32( (void*)tb[RTA_MP_ALGO]);
		    rta->have |= NIC_RT_MP_ALGO;
                    eprintf(NIC_DEBUG, "nic route: mp algorithm: %x", rta->mp_algo);
		}
		
		if ( tb[RTA_METRICS] )
		{
		    struct nlattr *rtax_tb [ __RTAX_MAX ];
		    int m;
		    for(m=0; m < __RTAX_MAX; m++) rtax_tb[m] = 0L;

		    err = nla_parse_nested( rtax_tb, RTAX_MAX, (void*)tb[RTA_METRICS], 0L);
		    if( err < 0 )
		    {
                        eprintf(NIC_ERR, "nic route: parse of RTAX metrics failed - %d %s",
			        nl_get_errno(), nl_geterror());
			/*XXX: do we care ?*/
		    }else
		    {
			for( m = 0; m < __RTAX_MAX ; m++ )
			{
			    if( rtax_tb [ m ] )
			    {
				rta->metrics[ m ] = nla_get_u32( (void*) rtax_tb [ m ] );
				rta->have_metrics |= 1<<(m+1);
			    }
			}
		    }
		}
		
		if ( tb[RTA_MULTIPATH] )
		{
		    size_t rtnh_len = nla_len((void*)tb[RTA_MULTIPATH]);

		    struct rtnexthop 
			*rtnhs = nla_data((void*)tb[RTA_MULTIPATH]), 
			*rtnhe = (struct rtnexthop*)(((unsigned long)rtnhs) + rtnh_len),
			*rtnh = rtnhs;
		    
		    while( rtnh < rtnhe )
		    {
			rta->n_hops += 1;
			rtnh = RTNH_NEXT(rtnh);
		    }
		    
		    rta->hops = calloc(rta->n_hops, sizeof(struct nic_rtnh_s));

		    if( rta->hops == 0L )
		    {
                        eprintf(NIC_ERR, "nic route: parse of route next hops failed - out of memory.");
			/*XXX: do we care ?*/
		    }else
		    {
			struct nic_rtnh_s *nic_rtnh = rta->hops;
			for(rtnh = rtnhs; rtnh < rtnhe; rtnh = RTNH_NEXT(rtnh), nic_rtnh++ )
			{
			    nic_rtnh->rtnh = *rtnh;
			    if( rtnh->rtnh_len > sizeof(struct rtnexthop) )
			    {
				struct nlattr *rtnhtb[ __RTA_MAX ];
				nla_parse((void*)rtnhtb, RTA_MAX, (struct nlattr *) RTNH_DATA(rtnh),
					  rtnh->rtnh_len - sizeof(struct rtnexthop), NULL
				         );				
				if ( rtnhtb[RTA_GATEWAY] ) 
				{
				    nic_rtnh->gw = 
					ip_addr_binary
					(   nla_data( (void*) rtnhtb[RTA_GATEWAY] ),
					    nla_len ( (void*) rtnhtb[RTA_GATEWAY] )
					);
				}
			    }
			}
		    }
		}

		n_routes++;

		RT_t  **prt=0, *trt=0L, *frt=rt, *nrt = 0L;

		if( rt == 0 )
		{
		    nrt = calloc(1, sizeof(struct nic_route_s));
		    if( nrt == 0L )
		    {
                        eprintf(NIC_ERR, "nic_get_routes: out of memory.");
			goto rta_return;
		    }
		    *nrt= *rtp;
                    eprintf(NIC_DEBUG," new route %p - rta: %p rtp->nh: %p", nrt, rta, nrt->nh);
		    trt =  rtp;
		    rtp =  nrt;
		    nrt =  trt;
		    trt =  0;
		}
		
		if( (nh->nic_route_tree == 0L)
		  ||((prt = tfind(rtp, &(nh->nic_route_tree), nic_route_comparator)) == 0L)
		  )
		{
		    tsearch(rtp, &(nh->nic_route_tree), nic_route_comparator);
                    eprintf(NIC_DEBUG," route tsearch %p - nh: %p", rtp, rtp->nh);
		}
		
		if( prt != 0L )
		{
		    trt = *prt;
		    *trt = *rtp;
		    frt =  trt;
		    eprintf(NIC_DEBUG," route found %p - freeing %p", trt, rtp);
		    if( nrt )
			free(rtp);
		}

		if( rt == 0L )
		    rtp = &srt;
		else
		if( nic_route_comparator( frt , rtp ) == 0 )
		    goto rta_return;
		
		hdr = nlmsg_next( hdr, &remaining );

	    } while ( remaining > 0 );	

	}else
	{
	    eprintf(NIC_ERR,"nic_get_routes: netlink recv failed: %s %d\n", 
		    nl_get_errno(), nl_geterror());
	    goto rta_return;
	}
    } while ( (hdr->nlmsg_type == RTM_NEWROUTE) && (remaining > 0) );

rta_return:
    if( buf )
	free(buf);
    nlmsg_free(msg);
    return n_routes;	    

nla_put_failure:
    nlmsg_free(msg);
    return n_routes;
}

struct nic_route_cbarg
{
    IProute_handler_t cb;
    void        *arg;
};

static void nic_route_twalker( const void *p, const VISIT which, const int depth )
{
    struct nic_route_s *route, *const*app=p;
    struct nic_route_cbarg *cb;

    if( (app == 0L)    || ((route = *app) == 0L)
      ||(route->nh == 0) || (route->nh->nic_route_foreach_arg == 0 )
      ||( (which != postorder) && (which != leaf) )
      ) return;

    cb = route->nh->nic_route_foreach_arg;
 
    cb->cb( route->nh, route, cb->arg );       
}

void nic_route_foreach(NLH_t nh, IProute_handler_t handler,  void *arg)
{
    struct nic_route_cbarg cb = { handler, arg };

    if ( nh->nic_route_foreach_arg )
	return;

    nh->nic_route_foreach_arg = &cb;

    nic_get_routes(nh, 0L);

    if( nh->nic_route_tree )
	twalk( nh->nic_route_tree, nic_route_twalker);

    nh->nic_route_foreach_arg = 0L;
}

int32_t nic_route_get_table(IProute_t rt)
{
    if( rt == 0L ) return 0;
    return rt->rt.rtm.rtm_table;
}

void nic_route_set_table(IProute_t rt, uint8_t table)
{
    if(rt == 0L) return;
    rt->rt.rtm.rtm_table = table;
}

char *nic_route_get_table_name( int32_t table, char *buf, int len )
{
    return rtnl_route_table2str( table, buf, len );
}

extern
int32_t nic_route_get_table_number( char *table )
{
    return rtnl_route_str2table( table );
}

uint8_t nic_route_get_family( IProute_t rt)
{
    if( rt == 0L ) return 0;
    
    return rt->rt.rtm.rtm_family;
}

uint8_t nic_route_get_scope(IProute_t rt)
{
    if( rt == 0L ) return 0;
    return rt->rt.rtm.rtm_scope;
}

void nic_route_set_scope(IProute_t rt, uint8_t scope)
{
    if( rt == 0L ) return;
    rt->rt.rtm.rtm_scope = scope;
}

uint32_t nic_route_get_flags( IProute_t rt )
{
    if( rt == 0L ) return 0;
    return rt->rt.rtm.rtm_flags;   
}

void nic_route_set_flags(IProute_t rt, uint32_t flags)
{
    if( rt == 0L ) return;
    rt->rt.rtm.rtm_flags = flags;
}

uint8_t nic_route_get_dst_len( IProute_t rt )
{
    if( rt == 0L ) return 0;
    return rt->rt.rtm.rtm_dst_len;   
}

void nic_route_set_dst_len( IProute_t rt, uint8_t dst_len )
{
    if( rt == 0L ) return;
    rt->rt.rtm.rtm_dst_len = dst_len;   
}

uint8_t nic_route_get_src_len( IProute_t rt )
{
    if( rt == 0L ) return 0;
    return rt->rt.rtm.rtm_src_len;   
}

void nic_route_set_src_len( IProute_t rt, uint8_t src_len )
{
    if( rt == 0L ) return;
    rt->rt.rtm.rtm_src_len = src_len;   
}

uint8_t nic_route_get_type( IProute_t rt)
{
    if( rt == 0L ) return 0;
    return rt->rt.rtm.rtm_type;
}

void nic_route_set_type( IProute_t rt, uint8_t type )
{    
    if( rt == 0L ) return;
    rt->rt.rtm.rtm_type = type;
}

uint8_t nic_route_get_protocol( IProute_t rt)
{
    if( rt == 0L ) return 0;
    return rt->rt.rtm.rtm_protocol;
}

void nic_route_set_protocol( IProute_t rt, uint8_t protocol )
{    
    if( rt == 0L ) return;
    rt->rt.rtm.rtm_protocol = protocol;
}

uint8_t nic_route_get_tos( IProute_t rt)
{
    if( rt == 0L ) return 0;
    return rt->rt.rtm.rtm_tos;
}

void nic_route_set_tos( IProute_t rt, uint8_t tos )
{    
    if( rt == 0L ) return;
    rt->rt.rtm.rtm_protocol = tos;
}

ip_addr_t nic_route_get_dst( IProute_t rt )
{
    ip_addr_t ip;
    memset(&ip,'\0',sizeof(ip));
    if( ( rt == 0L ) 
      ||((rt->rt.have & NIC_RT_DST)==0)
      ) return ip;
    return rt->rt.dst;
}

void nic_route_set_dst( IProute_t rt, ip_addr_t ip)
{
    if( rt == 0L ) return;
    rt->rt.dst = ip;
    rt->rt.have |= NIC_RT_DST;
}

ip_addr_t nic_route_get_gateway( IProute_t rt )
{
    ip_addr_t ip;
    memset(&ip,'\0',sizeof(ip));
    if( ( rt == 0L ) 
      ||((rt->rt.have & NIC_RT_GATEWAY)==0)
      ) return ip;
    return rt->rt.gw;
}

void nic_route_set_gateway( IProute_t rt, ip_addr_t ip)
{
    if( rt == 0L ) return;
    rt->rt.gw = ip;
    rt->rt.have |= NIC_RT_GATEWAY;
}

ip_addr_t nic_route_get_src( IProute_t rt )
{
    ip_addr_t ip;
    memset(&ip,'\0',sizeof(ip));
    if( ( rt == 0L ) 
      ||((rt->rt.have & NIC_RT_SRC)==0)
      ) return ip;
    return rt->rt.src;
}

void nic_route_set_src( IProute_t rt, ip_addr_t ip)
{
    if( rt == 0L ) return;
    rt->rt.src = ip;
    rt->rt.have |= NIC_RT_SRC;
}

ip_addr_t nic_route_get_prefsrc( IProute_t rt )
{
    ip_addr_t ip;
    memset(&ip,'\0',sizeof(ip));
    if( ( rt == 0L ) 
      ||((rt->rt.have & NIC_RT_PREFSRC)==0)
      ) return ip;
    return rt->rt.prefsrc;
}

void nic_route_set_prefsrc( IProute_t rt, ip_addr_t ip)
{
    if( rt == 0L ) return;
    rt->rt.prefsrc = ip;
    rt->rt.have |= NIC_RT_PREFSRC;
}

int16_t nic_route_get_oif( IProute_t rt )
{
    if( ( rt == 0L ) 
      ||((rt->rt.have & NIC_RT_OIF)==0)
      ) return -1;
    
    return rt->rt.oif;
}

void nic_route_set_oif( IProute_t rt, uint16_t oif)
{
    if( rt == 0L ) return ;
    
    rt->rt.oif = oif;
    rt->rt.have |= NIC_RT_OIF;
}

NIC_if_name_t nic_route_get_iif( IProute_t rt )
{
    NIC_if_name_t nil = { { '\0' } };
    if( ( rt == 0L )
      ||((rt->rt.have & NIC_RT_IIF) == 0 )
      ) return nil;
    
    return *((NIC_if_name_t*)&(rt->rt.iif[0]));
}

void nic_route_set_iif( IProute_t rt, char *iif )
{
    if( rt == 0 ) return;
    
    strncpy(&(rt->rt.iif[0]), iif, IFNAMSIZ);	
    rt->rt.have |= NIC_RT_IIF;
}

uint32_t nic_route_get_priority( IProute_t rt )
{
    if( ( rt == 0L )
      ||((rt->rt.have & NIC_RT_PRIORITY) == 0 )
      ) return 0;
    
    return rt->rt.priority;
}

void nic_route_set_priority( IProute_t rt, uint32_t priority)
{
    if( rt == 0L ) return;
    
    rt->rt.priority = priority;
    rt->rt.have |= NIC_RT_PRIORITY;
}

extern
uint32_t nic_route_get_protoinfo(IProute_t rt)
{
    if( ( rt == 0L )
      ||((rt->rt.have & NIC_RT_PROTOINFO) == 0 )
      ) return 0;
    
    return rt->rt.protoinfo;
}

extern
void nic_route_set_protoinfo(IProute_t rt, uint32_t protoinfo)
{
    if( rt == 0L ) return;
    rt->rt.protoinfo = protoinfo;
    rt->rt.have |= NIC_RT_PROTOINFO;
}

uint32_t nic_route_get_session(IProute_t rt)
{
    if( ( rt == 0L )
      ||((rt->rt.have & NIC_RT_SESSION) == 0 )
      ) return 0;
    
    return rt->rt.session;
}

extern
void nic_route_set_session(IProute_t rt, uint32_t session)
{
    if( rt == 0L ) return;
    rt->rt.session = session;
    rt->rt.have |= NIC_RT_SESSION;
}

uint32_t nic_route_get_flow(IProute_t rt)
{
    if( ( rt == 0L )
      ||((rt->rt.have & NIC_RT_FLOW) == 0 )
      ) return 0;
    
    return rt->rt.flow;
}

void nic_route_set_flow(IProute_t rt, uint32_t flow)
{
    if( rt == 0L ) return;
    rt->rt.flow = flow;
    rt->rt.have |= NIC_RT_FLOW;
}

uint32_t nic_route_get_mp_algo(IProute_t rt)
{
    if( ( rt == 0L )
      ||((rt->rt.have & NIC_RT_MP_ALGO) == 0 )
      ) return 0;
    
    return rt->rt.mp_algo;
}

void nic_route_set_mp_algo(IProute_t rt, uint32_t mp_algo)
{
    if( rt == 0L ) return;
    rt->rt.mp_algo = mp_algo;
    rt->rt.have |= NIC_RT_MP_ALGO;
}

IProute_t nic_route_new
(   NLH_t nh,
    int16_t oif,          /* output interface: -1 means none */
    ip_addr_t *dst,       /* 0L here means DEFAULT ROUTE */ 
    uint8_t dst_len,      /* dst_len ignored for default route */
    ip_addr_t *gw,        /* 0L here means no gateway */
    int8_t type,          /* -1: default: RTN_UNICAST */
    int8_t protocol,      /* -1: default: RTPROT_BOOT */
    int8_t scope,         /* -1=UNIVERSE,LINK=253,SITE=200,HOST=254 */
    int32_t priority,     /* metric - -1 means no priority */
    int32_t table,        /* table: -1 selects default: "local" */
    char*   iif,          /* input interface: 0 means none */
    ip_addr_t *src,       /* 0L here means no src route */
    uint8_t  src_len      /* 0 here means no src route */
)
{
    if( ( dst && (dst->sa_family != AF_INET) && (dst->sa_family != AF_INET6))
      ||( gw && (gw->sa_family != AF_INET) && (gw->sa_family != AF_INET6))
      ||( src && (src->sa_family != AF_INET) && (src->sa_family != AF_INET6))
      )
	return 0L;
	
    IProute_t rt = calloc(1, sizeof(struct nic_route_s));

    if(rt == 0L)
	return rt;

    rt->nh = nh;


    if ( dst == 0L )
    { /* default route */
	dst_len = 0;
	table = RT_TABLE_MAIN;
    }else
    {
	rt->rt.rtm.rtm_dst_len = dst_len;
	rt->rt.dst = *dst;
	rt->rt.rtm.rtm_family = dst->sa_family;
	rt->rt.have |= NIC_RT_DST;
    }

    if( gw  )
    {	
	rt->rt.gw = *gw;
	rt->rt.have |= NIC_RT_GATEWAY;
	if( rt->rt.rtm.rtm_family == 0 )
	    rt->rt.rtm.rtm_family = rt->rt.gw.sa_family;
    }

    if( scope == -1 )
	rt->rt.rtm.rtm_scope = RT_SCOPE_UNIVERSE;
    else
	rt->rt.rtm.rtm_scope = scope;
    
    if( table == -1 )
	rt->rt.rtm.rtm_table = RT_TABLE_LOCAL;
    else
	rt->rt.rtm.rtm_table = table;

    if( type == -1 )
	rt->rt.rtm.rtm_type = RTN_UNICAST;

    if( protocol == -1 )
	rt->rt.rtm.rtm_protocol = RTPROT_BOOT;

    if( oif > -1 )
    {
	rt->rt.oif = oif;
	rt->rt.have |= NIC_RT_OIF;
    }
   
    if( iif )
    {
	strncpy(&(rt->rt.iif[0]), iif, IFNAMSIZ);
	rt->rt.have |= NIC_RT_IIF;
    }

    if( src && src_len )
    {	
	rt->rt.src = *src;
	rt->rt.have |= NIC_RT_SRC;
	if( rt->rt.rtm.rtm_family == 0 )
	    rt->rt.rtm.rtm_family = rt->rt.src.sa_family;
    }

    if( priority > -1 )
    {
	rt->rt.priority = priority;
	rt->rt.have |= NIC_RT_PRIORITY;
    }
    return rt;
}

static NIC_Res_t nic_route_msg( IProute_t rt, char *fn, char *tag, uint32_t msg_type, uint32_t msg_flags)
{
    NLH_t nh = rt->nh;
    /* libnl as yet provides no support for route creation / deletion / modification.
     */
    struct nl_msg *msg = nlmsg_build_simple(msg_type, msg_flags) ;
    if( msg == 0L )
    {
	eprintf(NIC_ERR,"%s: failed to %s route - out of memory.", tag, fn);
	return NIC_FAIL;
    }
    
    if ( nlmsg_append(msg, &(rt->rt.rtm), sizeof(struct rtmsg), 1) < 0 )
	goto nla_put_failure;

    if( rt->rt.have & NIC_RT_OIF )
	NLA_PUT_U32(msg, RTA_OIF, rt->rt.oif);
    
    if ( rt->rt.have & NIC_RT_DST )
	NLA_PUT(msg, RTA_DST, IP_ADDR_SIZE(&(rt->rt.dst)), IP_ADDR(&(rt->rt.dst)));

    if ( rt->rt.have & NIC_RT_GATEWAY )
	NLA_PUT(msg, RTA_GATEWAY, IP_ADDR_SIZE(&(rt->rt.gw)), IP_ADDR(&(rt->rt.gw)));

    if( rt->rt.have & NIC_RT_IIF )
	NLA_PUT_STRING(msg, RTA_IIF, &(rt->rt.iif[0]));

    if ( rt->rt.have & NIC_RT_SRC )
	NLA_PUT(msg, RTA_SRC, IP_ADDR_SIZE(&(rt->rt.src)), IP_ADDR(&(rt->rt.src)));

    if ( rt->rt.have & NIC_RT_PREFSRC )
	NLA_PUT(msg, RTA_SRC, IP_ADDR_SIZE(&(rt->rt.prefsrc)), IP_ADDR(&(rt->rt.prefsrc)));

    if ( rt->rt.have & NIC_RT_PRIORITY )
	NLA_PUT_U32(msg, RTA_PRIORITY, rt->rt.priority);
    
    if ( rt->rt.have & NIC_RT_FLOW )
	NLA_PUT_U32(msg, RTA_FLOW, rt->rt.flow);
    
    if ( rt->rt.have & NIC_RT_SESSION )
	NLA_PUT_U32(msg, RTA_SESSION, rt->rt.session);

    if ( rt->rt.have & NIC_RT_MP_ALGO )
	NLA_PUT_U32(msg, RTA_MP_ALGO, rt->rt.mp_algo);
    
    if( nl_send_auto_complete(rt->nh->nl, nlmsg_hdr(msg)) < 0 )
    {
	eprintf(NIC_ERR,"%s: send failed - %d %s", tag, nl_get_errno(),
                nl_geterror());
	return NIC_FAIL;
    }

    nlmsg_free(msg);

    if( nl_wait_for_ack(rt->nh->nl) < 0 )
    {
        eprintf(NIC_ERR,"%s: failed - %d: %s", tag, nl_get_errno(),
                nl_geterror());
	return NIC_FAIL;	
    }
    
    return NIC_OK;

nla_put_failure:

    eprintf(NIC_ERR,"%s: failed to %s route - out of memory.", tag, fn);
    nlmsg_free(msg);
    return NIC_FAIL;
}

NIC_Res_t nic_add_route( IProute_t rt )
{ 
    if (rt == 0L) return NIC_FAIL;

    return nic_route_msg( rt, "nic_add_route", "add", RTM_NEWROUTE, NLM_F_CREATE | NLM_F_REPLACE);
}

NIC_Res_t nic_update_route( IProute_t rt )
{ 
    if (rt == 0L) return NIC_FAIL;

    return nic_route_msg( rt, "nic_update_route", "update", RTM_NEWROUTE, NLM_F_REPLACE);
}

NIC_Res_t nic_remove_route( IProute_t rt )
{ 
    if (rt == 0L) return NIC_FAIL;

    return nic_route_msg( rt, "nic_remove_route", "remove", RTM_DELROUTE, 0);
}

IProute_list_t *nic_route_list_new( IProute_t rt1, ... )
{      
    va_list va;
    NLH_t nh = rt1->nh;
    va_start(va, rt1);
    
    IProute_list_t *head = calloc(1,sizeof(IProute_list_t));
    IProute_list_node_t *n;
    IProute_t rt;

    if( head == 0 )
    {
	va_end(va);
        eprintf(NIC_ERR, "out of memory" );
	return 0;
    }
    rt=rt1;
    STAILQ_INIT(head);
    do
    {
	n = calloc(1,sizeof(IProute_list_node_t));
	if ( n == 0L )
	{
	    va_end(va);
	    free(head);
            eprintf(NIC_ERR, "out of memory");
	    return 0L;
	}
	n->route = rt;
	STAILQ_INSERT_TAIL( head, n, link );
    }while( (rt = va_arg(va, IProute_t)) != 0) ;
    va_end(va);
    return(head);
}

void nic_route_list_free( IProute_list_t *head )
{
    IProute_list_node_t *n, *a;
    
    if ( !STAILQ_EMPTY(head) )
    {
	n = STAILQ_FIRST( head );
	do
	{
	    a = STAILQ_NEXT(n, link);
	    nic_route_free(n->route);
	    free(n);
	    n = a;
	}while (n);
    }
}

NIC_Res_t nic_add_routes( IProute_list_t *head )
{
    IProute_list_node_t *n;
    NIC_Res_t r;
    STAILQ_FOREACH(n,head,link)
    {
	if((r=nic_add_route(n->route)) == NIC_FAIL)
	    return r;
    }
    return NIC_OK;
}

NIC_Res_t nic_remove_routes( IProute_list_t *head )
{
    IProute_list_node_t *n;
    NIC_Res_t r;
    STAILQ_FOREACH(n,head,link)
    {
	if((r=nic_remove_route(n->route))==NIC_FAIL)
	    return r;
    }
    return NIC_OK;
}

NIC_Res_t nic_update_routes( IProute_list_t *head )
{
    IProute_list_node_t *n;
    NIC_Res_t r;
    STAILQ_FOREACH(n,head,link)
    {
	if((r=nic_update_route(n->route))==NIC_FAIL)
	    return r;
    }
    return NIC_OK;
}

void nic_route_free(void *rtp)
{
    IProute_t rt = rtp;

    if (!rt)
        return;
    
    if( rt->nh->nic_route_tree )
	tdelete(rt, &(rt->nh->nic_route_tree), nic_route_comparator);
    
    if( rt->rt.hops )
	free(rt->rt.hops);

    free(rt);
}

extern
NIC_Res_t 
nic_set_hostname
( NLH_t nh,
  char *hostname
)
{
    return sethostname(hostname, strlen(hostname)) == 0 ? NIC_OK : NIC_FAIL;
}

NIC_Res_t 
nic_configure
(
    NLH_t          nh,
    NIC_t          nic,
    IPaddr_list_t  *addresses,
    IProute_list_t *routes,
    IPaddr_list_t  *dns_servers,
    char           *search_list,
    char           *host_name
)
{
    if ( addresses && ( nic_add_addresses( nic, addresses ) == NIC_FAIL ))
    {
        eprintf(NIC_ERR, "nic_configure: failed to add addresses.");
	return NIC_FAIL;
    }
     
    if ( routes && ( nic_add_routes( routes ) == NIC_FAIL ))
    {
        eprintf(NIC_ERR, "nic_configure: failed to add routes.");
	return NIC_FAIL;
    }

    if ( host_name && ( nic_set_hostname( nh, host_name ) == NIC_FAIL ))
    {
        eprintf(NIC_ERR, "nic_configure: failed to set hostname.");
	return NIC_FAIL;
    }

    return NIC_OK;
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
