/* pump.c
 *
 * libdhcp pump compatibility API for anaconda -
 * only the pump symbols used by anaconda are defined.
 *
 ******************************************************************************
 * Original pump credits:
 * 
 * There are a bunch of folks who made this code possible. I'm sure I've missed
 * some in this list :-(
 *
 *  Alan Cox <alan@redhat.com>
 *  Bruce Beare <bbeare@cisco.com>
 *  David Blythe <blythe@routefree.com>
 *  Stephen Carville <carville@cpl.net>
 *  Guy Delamarter <delamart@pas.rochester.edu>
 *  Chris Johnson <cjohnson@mint.net>
 *  Michael Johnson <johnsonm@redhat.com>
 *  H. J. Lu <hjl@valinux.com>
 *  Kristof Petr <Petr@Kristof.CZ>
 *  Marco Pietrobono <pietrobo@pietrobo.com>
 *  Benjamin Reed <breed@cse.ucsc.edu>
 *  George Staikos <staikos@0wned.org>
 *  Jay Turner <jturner@redhat.com>
 *  Matt Wilson <msw@redhat.com>
 *  <aaron@schrab.com>
 *  <duanev@io.com>
 *  <dunham@cse.msu.edu>
 *  <safford@watson.ibm.com>
 *  weejock@ferret.lmh.ox.ac.uk
 *
 *******************************************************************************
 *
 * Copyright 1999-2001 Red Hat, Inc.
 * 
 * All Rights Reserved.
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * OPEN GROUP BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of Red Hat shall not be
 * used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Red Hat.
 *
 ******************************************************************************
 *
 * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of
 * the GNU General Public License v.2.  This program is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY expressed or
 * implied, including the implied warranties of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.  You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA. Any Red Hat trademarks that are incorporated in the source code or
 * documentation are not subject to the GNU General Public License and may
 * only be used or replicated with the express permission of Red Hat, Inc.
 *
 * Red Hat Author(s): Jason Vas Dias
 *                    David Cantrell
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/queue.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <pump.h>
#include <netlink/netlink.h>
#include <netlink/netlink-kernel.h>

#include <dhcp4client/dhcp4client.h>
#include <dhcp4client/isc_dhcp/dhcpd.h>

/* DHCP options to request */
#define OPTLIST "subnet-mask,broadcast-address,routers,domain-name,domain-name-servers,host-name,static-routes"

enum {
    OOM,
    DHCP_FAILED,
};

static char * const pump_errors[] = {
    [OOM] = "out of memory",
    [DHCP_FAILED] = "DHCP configuration failed",
    NULL
};

char * pumpSetupInterface(struct pumpNetIntf * ifx) {
    static char buf[512];
    NIC_t nic = 0;
    IPaddr_t addr = 0;
    IProute_t route = 0;
    int  err=0;
    char *tag="";

    if (ifx->dhcp_nic) {
        /* do DHCP configuration */
        pumpEnableInterface(ifx->device);

        assert(ifx->dhcp_nic->nh);

        int r = dhcp_nic_configure(ifx->dhcp_nic);
	
        dhcp_nic_free(ifx->dhcp_nic);
        ifx->dhcp_nic = 0L;
        if (r < 0) {
            snprintf(buf, 512, "DHCP configuration failed - %d %s.",
                     -r, strerror(-r));
            relinquish_timeouts();
            return buf;
        }

        relinquish_timeouts();
        return NULL;
    }

    /* do anaconda static configuration */
    if (!ifx->nh) {
        ifx->nh = nic_open(0);
        if (!ifx->nh) {
            err = ENOMEM;
            tag = "nic_open";
            goto return_emsg;
        }
    }

    nic = nic_by_name(ifx->nh, ifx->device);
    if (nic == 0) {
        err = ENODEV;
        tag = "get link";
        nic_close(&ifx->nh);
        goto return_emsg;
    }

    uint32_t flags = nic_get_flags(nic);

    if (flags & (IFF_UP | IFF_RUNNING | IFF_BROADCAST))
        nic_set_flags(nic, flags | IFF_UP | IFF_RUNNING | IFF_BROADCAST);

    if ((ifx->set & PUMP_INTFINFO_HAS_MTU) && (ifx->mtu))
        nic_set_mtu(nic, ifx->mtu);

    tag = "set link";
    if ((err = - nic_update(nic)) > 0) {
        nic_close(&ifx->nh);
        goto return_emsg;
    }

    if ((ifx->ip.sa_family != AF_INET) && (ifx->ip.sa_family != AF_INET6)) {
        nic_close(&ifx->nh);
        return NULL;
    }

    addr = nic_addr(ifx->nh, ifx->ip);
    uint8_t prefix = 0;
    uint8_t isIPv4 = 0;

    if (addr == 0L) {
        err = ENOMEM;
        tag = "new addr";
        nic_close(&ifx->nh);
        goto return_emsg;
    }

    if (ifx->netmask.sa_family)
        switch (ifx->ip.sa_family) {
            case AF_INET:
                isIPv4 = 1;

                if (ip_v4_addr(&(ifx->netmask))) {
                    prefix = ip_v4_netmask_to_prefix(&(ifx->netmask));
                    nic_addr_set_prefix(addr, prefix);
                }

                break;
            case AF_INET6:
                if (ifx->ipv6_prefixlen) {
                    nic_addr_set_prefix(addr, ifx->ipv6_prefixlen);
                }

                break;
        }

    if (ifx->broadcast.sa_family) {
        nic_addr_set_broadcast(addr, ifx->broadcast);
    } else if (prefix && isIPv4) {
        nic_addr_set_broadcast(addr, ip_v4_broadcast(&(ifx->ip), prefix));
    }

    tag = "create addr";

    if ((err = -nic_add_address(nic, addr)) > 0) {
        nic_addr_free(addr);
        nic_close(&ifx->nh);
        goto return_emsg;
    }
    
    /* XXX: do we still need to support pre-2.2 kernels ?
     * if so, we may need to add a network route here ...
     */

    route = 0;

    if (ifx->gateway.sa_family) {
        tag = "new route";
        route = nic_route_new(ifx->nh,
                              nic_get_index(nic), /* output interface oif     */
                              0L,0,               /* dst=0, len=0: DEFAULT !  */
                              &ifx->gateway,      /* gateway ip_addr_t        */
                              -1,                 /* -1: default: RTN_UNICAST */
                              -1,                 /* -1: default: RTPROT_BOOT */
                              -1,                 /* default scope: UNIVERSE  */
                              -1,                 /* no priority              */
                              -1,                 /* table: default (main)    */
                              0L,                 /* no input interface       */
                              0L,                 /* no src                   */
                              0);                 /* no src len               */

        if (route == 0) {
            err = ENOMEM;
            nic_addr_free(addr);
            nic_close(&ifx->nh);
            goto return_emsg;
        }

        tag = "create route";

        if ((err = -nic_add_route(route)) > 0) {
            nic_route_free(route);
            nic_addr_free(addr);
            nic_close(&ifx->nh);
            goto return_emsg;
        }
    }

    if (route != 0)
        nic_route_free(route);

    nic_addr_free(addr);
    nic_close(&ifx->nh);
    return NULL;

return_emsg:
    snprintf(buf, 512, "pumpSetupInterface failed: %s - %d: %s.",
             tag, err, strerror(err));
    return buf;
}

static void ifxDHCPv6 (struct pumpNetIntf * ifx) {
    int r;
    IPaddr_t addr = STAILQ_FIRST(&(ifx->dhcp_nic->dhcpv6->address_list))->addr;

    if (ifx->set & PUMP_INTFINFO_HAS_IP) {
        ifx->ipv6 = nic_ip_addr(addr);

        if ((ifx->ipv6_prefixlen = nic_addr_get_prefix(addr)) > 0)
            ifx->set |= PUMP_INTFINFO_HAS_IPV6_PREFIX;

        ifx->set |= PUMP_INTFINFO_HAS_IPV6_IP;
    } else {
        ifx->ip = nic_ip_addr(addr);
        ifx->set |= PUMP_INTFINFO_HAS_IP;
        ifx->ipv6_prefixlen = nic_addr_get_prefix(addr);

        if (ifx->ipv6_prefixlen) {
            ifx->network = ip_mask(&ifx->ip, ifx->ipv6_prefixlen);
            ifx->set |= PUMP_INTFINFO_HAS_NETWORK;
        }
    }

    if (!STAILQ_EMPTY(&(ifx->dhcp_nic->dhcpv6->dns_list))) {
        IPaddr_list_node_t *n = 0;
        STAILQ_FOREACH(n, &(ifx->dhcp_nic->dhcpv6->dns_list), link) {
            if (ifx->numDns >= MAX_DNS_SERVERS)
                break;

            ifx->dnsServers[ifx->numDns++] = nic_ip_addr(n->addr);
            ifx->set |= PUMP_NETINFO_HAS_DNS;
        }
    }

    if ((ifx->set & PUMP_NETINFO_HAS_DNS) &&
        ifx->dhcp_nic->dhcpv6->search_list) {
        char *s = ifx->domain;

        r = asprintf(&(ifx->domain), "%s%s%s", s ? s : "", s ? " " : "",
                     ifx->dhcp_nic->dhcpv6->search_list);

        if (s)
            free(s);

        ifx->set |= PUMP_NETINFO_HAS_DOMAIN;
    }
}

static void ifxDHCPv4(struct pumpNetIntf *ifx) {
    int r;
    IPaddr_t addr = STAILQ_FIRST(&(ifx->dhcp_nic->dhcpv4->address_list))->addr;

    if (ifx->set & PUMP_INTFINFO_HAS_IP) {
        ifx->ipv4 = nic_ip_addr(addr);
        ifx->set |= PUMP_INTFINFO_HAS_IPV4_IP;
    } else {
        ifx->ip = nic_ip_addr(addr);
        ifx->set |= PUMP_INTFINFO_HAS_IP;

        if (nic_addr_get_prefix(addr)) {
            ifx->network = ip_mask(&ifx->ip, nic_addr_get_prefix(addr));
            ifx->set |= PUMP_INTFINFO_HAS_NETWORK;
        }
    }

    if (nic_addr_get_prefix(addr)) {
        ifx->netmask = ip_v4_prefix_to_netmask(nic_addr_get_prefix(addr));
        ifx->set |= PUMP_INTFINFO_HAS_NETMASK;
        ip_addr_t broadcast = nic_addr_get_broadcast(addr);

        if (broadcast.sa_family) {
            ifx->broadcast = broadcast;
            ifx->set |= PUMP_INTFINFO_HAS_BROADCAST;
        }		
    }

    if (ifx->dhcp_nic->dhcpv4 &&
        (!STAILQ_EMPTY(&(ifx->dhcp_nic->dhcpv4->route_list)))) {
        IProute_list_node_t *n;

        STAILQ_FOREACH(n, &(ifx->dhcp_nic->dhcpv4->route_list), link) {
            ip_addr_t dst = nic_route_get_dst(n->route);
            ip_addr_t gw = nic_route_get_gateway(n->route);
            uint8_t dst_len = nic_route_get_dst_len(n->route);

            if ((dst.sa_family == 0) && (dst_len == 0) &&
                (gw.sa_family == AF_INET)) {
                ifx->gateway = gw;
                ifx->set |= PUMP_NETINFO_HAS_GATEWAY;
                break;
            }		
        }
    }

    if (ifx->dhcp_nic->dhc4ctl && dhcpv4_mtu_option(ifx->dhcp_nic->dhc4ctl)) {
        ifx->mtu =  dhcpv4_mtu_option(ifx->dhcp_nic->dhc4ctl);
        ifx->set |= PUMP_INTFINFO_HAS_MTU;
    }

    if (ifx->dhcp_nic->dhcpv4->lease.dhcp4_lease) {
        DHCPv4_lease *lease=ifx->dhcp_nic->dhcpv4->lease.dhcp4_lease;

        if (lease->server_address.s_addr != 0) {
            ifx->nextServer = ip_addr_in(&(lease->server_address));
            ifx->set |= PUMP_INTFINFO_HAS_NEXTSERVER;
        }

        if (lease->filename) {
            ifx->bootFile = strdup(lease->filename);
            ifx->set |= PUMP_INTFINFO_HAS_BOOTFILE;
        }
    }

    if (!STAILQ_EMPTY(&(ifx->dhcp_nic->dhcpv4->dns_list))) {
        IPaddr_list_node_t *n = 0;

        STAILQ_FOREACH(n, &(ifx->dhcp_nic->dhcpv4->dns_list), link) {
            if (ifx->numDns >= MAX_DNS_SERVERS)
                break;

            ifx->dnsServers[ifx->numDns++] = nic_ip_addr(n->addr);
            ifx->set |= PUMP_NETINFO_HAS_DNS;
        }
    }

    if ((ifx->set & PUMP_NETINFO_HAS_DNS) &&
        ifx->dhcp_nic->dhcpv4->search_list) {
        char *s = ifx->domain;

        r = asprintf(&(ifx->domain), "%s%s%s", s ? s : "", s ? " " : "",
                     ifx->dhcp_nic->dhcpv4->search_list);

        if (s)
            free(s);

        ifx->set |= PUMP_NETINFO_HAS_DOMAIN;
    }   
 
    if (ifx->dhcp_nic->dhcpv4->host_name) {
        ifx->set |= PUMP_NETINFO_HAS_HOSTNAME;

        if (ifx->hostname)
            free(ifx->hostname);

        ifx->hostname = strdup(ifx->dhcp_nic->dhcpv4->host_name);
    }
}

char * pumpDhcpClassRun (struct pumpNetIntf * ifx, char * hostname,
                         char * vendor_class, DHCP_Preference dhcp_preference,
                         LIBDHCP_Capability dhcp_capability,
                         time_t timeout,
                         void (*logger)(void*,int,char*,va_list),
                         int log_level) {
    char *disret;
    disret = pumpDisableInterface(ifx->device);
    if (disret != NULL)
        return disret;

    assert(!ifx->nh);

    if (!(ifx->nh = nic_open(NETLINK_ROUTE)))
        return pump_errors[OOM];

    if (logger) {
        nic_set_va_logger(ifx->nh, (NIC_VA_Error_Handler_t)logger, &ifx);
        nic_set_loglevel(ifx->nh, log_level );
    }

    pumpEnableInterface(ifx->device);
    ifx->dhcp_nic = dhcp_nic(ifx->nh, dhcp_preference, ifx->device,
                             dhcp_capability, timeout,
                             (LIBDHCP_Error_Handler) logger, log_level,
                             "-R", OPTLIST,
                             "-V", vendor_class,
                             hostname ? "-H" : 0L, hostname ? hostname : 0L,
                             0L);
    if (ifx->dhcp_nic) {
        /* fill in the pumpNetIntf structure.
         * We'll just use the FIRST configured address as ifx->ip -
        * users should realize that a list of addresses may have
        * been configured.
        */	
        if ((ifx->dhcp_nic->dhcpv4 && ifx->dhcp_nic->dhcpv6) && 
            (!STAILQ_EMPTY(&(ifx->dhcp_nic->dhcpv4->address_list))) &&
            (!STAILQ_EMPTY(&(ifx->dhcp_nic->dhcpv6->address_list)))) {
            if (dhcp_preference & IPv6_PREFERENCE) {
                ifxDHCPv6(ifx);
                ifxDHCPv4(ifx);
            } else {
                ifxDHCPv4(ifx);
                ifxDHCPv6(ifx);
            }
        } else if (ifx->dhcp_nic->dhcpv4 &&
                   (!STAILQ_EMPTY(&(ifx->dhcp_nic->dhcpv4->address_list)))) {
            ifxDHCPv4(ifx);
        } else if (ifx->dhcp_nic->dhcpv6 &&
                   (!STAILQ_EMPTY(&(ifx->dhcp_nic->dhcpv6->address_list)))) {
            ifxDHCPv6(ifx);
        }
	
        ifx->nh = NULL;
    } else {
        nic_close(&ifx->nh);
        return pump_errors[DHCP_FAILED];
    }

    return NULL;
}

char * pumpDisableInterface(char * device) {
    static char buf[512];
    struct ifreq req;
    int s, e;

    s = socket(AF_INET, SOCK_DGRAM, 0);

    memset(&req, 0, sizeof(req));
    strcpy(req.ifr_name, device);

    if (ioctl(s, SIOCGIFFLAGS, &req)) {
        e = errno;
        snprintf(buf, 512, "ioctl SIOCGIFFLAGS failed: %d %s\n",e,strerror(e));
        close(s);
        return &(buf[0]);
    }

    req.ifr_flags &= ~(IFF_UP | IFF_RUNNING);
    if (ioctl(s, SIOCSIFFLAGS, &req)) {
        e = errno;
        snprintf(buf, 512, "ioctl SIOCSIFFLAGS failed: %d %s\n",e,strerror(e));
        close(s);
        return &(buf[0]);
    }

    close(s);

    return NULL;
}

char * pumpEnableInterface(char * device) {
    static char buf[512];
    struct ifreq req;
    int s, e;

    s = socket(AF_INET, SOCK_DGRAM, 0);

    memset(&req, 0, sizeof(req));
    strcpy(req.ifr_name, device);

    if (ioctl(s, SIOCGIFFLAGS, &req)) {
        e = errno;
        snprintf(buf, 512, "ioctl SIOCGIFFLAGS failed: %d %s\n",e,strerror(e));
        close(s);
        return &(buf[0]);
    }

    req.ifr_flags |= (IFF_UP | IFF_RUNNING);
    if (ioctl(s, SIOCSIFFLAGS, &req)) {
        e = errno;
        snprintf(buf, 512, "ioctl SIOCSIFFLAGS failed: %d %s\n",e,strerror(e));
        close(s);
        return &(buf[0]);
    }

    close(s);

    return NULL;
}

/*
 * vim:ts=8:sw=4:sts=4:et
 */
